import torch
from mgp.model.data.dataloader import dataloader


class customLoader(object):
    def __init__(
        self,
        X,
        y,
        ind_total_nan,
        ind_total_nan_target,
        minibatch_size,
        numThreads=1,
        shuffle=True,
        device=None,
    ):
        self.dataset = dataloader()
        self.dataset.create(X, y, ind_total_nan, ind_total_nan_target, device=device)
        # load data set
        self.dataloader = torch.utils.data.DataLoader(
            self.dataset,
            batch_size=minibatch_size,
            shuffle=shuffle,
            # num_workers=numThreads
        )

        self.initialize()

    def name(self):
        return "customLoader"

    def initialize(self):
        pass

    def __len__(self):
        return len(self.dataset)

    def __iter__(self):
        for i, data in enumerate(self.dataloader):
            yield data



import torch
import numpy as np

default_jitter = 1e-7
default_float = torch.float64
import os

rmse_metric = torch.nn.MSELoss()


def compute_nn(y_t, mean_t, std_t):
    f1 = torch.log(std_t**2) / 2
    f2 = (y_t - mean_t) ** 2 / (2 * std_t**2)
    f3 = np.log(2 * np.pi) / 2
    # print(std_t.mean())
    return torch.mean(f1 + f2 + f3)


def compute_nn_mse(y, mean_pred, std_pred, ind_nan_target):
    D = ind_nan_target.shape[1]
    nll, rmse = [], []
    for d in range(D):
        if ind_nan_target[:, d].sum() > 0:
            a = rmse_metric(
                y[ind_nan_target[:, d], d],
                mean_pred[:, ind_nan_target[:, d], d].mean(0),
            ).sqrt()
            b = compute_nn(
                y[ind_nan_target[:, d], d],
                mean_pred[:, ind_nan_target[:, d], d].squeeze(-1),
                std_pred[:, ind_nan_target[:, d], d].squeeze(-1),
            )

            nll.append(b)
            rmse.append(a)

    nll = torch.stack(nll)
    rmse = torch.stack(rmse)
    # return nll[~nll.isnan()].sum(), rmse[~rmse.isnan()].sum()
    return nll.mean(), rmse.mean()


def compute_metrics(y, mean_pred, std_pred, ind_nan):
    def compute_nn0(y_t, mean_t, std_t):
        S = mean_t.shape[0]
        normal = torch.distributions.Normal(loc=mean_t, scale=std_t)
        logpdf = normal.log_prob(y_t)
        nll = torch.logsumexp(logpdf, 0) - np.log(S)
        return -nll.mean()

    def compute_nn(y_t, mean_t, std_t):
        # std_t = std_t/std_t.max()
        f1 = torch.log(std_t**2) / 2
        f2 = (y_t - mean_t) ** 2 / (2 * std_t**2)
        f3 = np.log(2 * np.pi) / 2

        return torch.mean(f1 + f2 + f3)

    if ind_nan is not None:
        mean_test = mean_pred[ind_nan]
        std_test = std_pred[ind_nan]
        y_test = y[ind_nan]

        mean_train = mean_pred[~ind_nan]
        std_train = std_pred[~ind_nan]
        y_train = y[~ind_nan]

        nll_test = compute_nn(y_test, mean_test, std_test)
        rmse_test = rmse_metric(y_test, mean_test).sqrt()

        nll_train = compute_nn(y_train, mean_train, std_train)
        rmse_train = rmse_metric(y_train, mean_train).sqrt()
        return rmse_train, nll_train, rmse_test, nll_test
    else:
        nll_test = compute_nn(y, mean_pred, std_pred)
        rmse_test = rmse_metric(y, mean_pred).sqrt()
        return rmse_test, nll_test


def accuracy(out, labels):
    if out.min() < 0:
        outputs = torch.where(out < 0.5, -1.0, 1.0)
    else:
        outputs = torch.where(out < 0.5, 0.0, 1.0)
    return torch.sum(outputs == labels) / (labels.shape[0])


# save models to the disk
def save_model(model, optimizer, device, model_path, epoch="latest"):
    save_filename = "VSGP_%s.pth" % (epoch)
    save_path = os.path.join(model_path, "checkpoints/")
    if not os.path.exists(save_path):
        os.mkdir(save_path)
    save_filepath = os.path.join(model_path, "checkpoints/", save_filename)
    np_state = np.random.get_state()
    if device.type == "cuda" and torch.cuda.is_available():
        trch_rng_state_cuda = torch.cuda.get_rng_state(device).cpu()
        trch_rng_state = torch.get_rng_state().cpu()
        state = {
            "state_dict": model.cpu().state_dict(),
            "optimizer": optimizer.state_dict(),
            "np_state": np_state,
            "trch_state": trch_rng_state,
            "trch_cuda_state": trch_rng_state_cuda,
        }
        torch.save(state, save_filepath)
        model.cuda(device)
    else:
        # trch_rng_state = torch.get_rng_state(device)
        state = {
            "state_dict": model.cpu().state_dict(),
            "optimizer": optimizer.state_dict(),
            "np_state": np_state,
        }
        torch.save(state, save_filepath)


def load_model(model, optimizer, model_path, device, which_epoch="latest"):
    # load models from the disk
    load_filename = "VSGP_%s.pth" % (which_epoch)

    load_path = os.path.join(model_path, "checkpoints/", load_filename)

    print("loading the model from %s" % load_path)
    # if you are using PyTorch newer than 0.4 (e.g., built from
    # GitHub source), you can remove str() on self.device
    state_dict = torch.load(load_path, map_location=device)

    if "state_dict" in state_dict:
        model.load_state_dict(state_dict["state_dict"])
        # optimizer.load_state_dict(state_dict['optimizer'])
        if torch.cuda.is_available() and device.type == "cuda":
            torch.cuda.set_rng_state(state_dict["trch_cuda_state"].cpu())
            torch.set_rng_state(state_dict["trch_state"].cpu())
        np.random.set_state(state_dict["np_state"])
        for state in optimizer.state.values():
            for k, v in state.items():
                if isinstance(v, torch.Tensor):
                    state[k] = v.to(device)
                else:
                    state[k] = v

    else:
        model.load_state_dict(state_dict)

    model.eval()


def reparameterize(mean, var, z, full_cov=False):
    """
    The 'reparameterization trick' for the Gaussian
    """

    if var is None:
        return mean

    if full_cov is False:
        return mean + z * (var + default_jitter) ** 0.5



import torch

import numpy as np

import warnings
from scipy.cluster.vq import kmeans2
import sys

from mgp.model.data import customLoader
from mgp.model.data.preprocess import load_dataset
from mgp.model.kernels import SquaredExponential as RBF
from mgp.model.kernels import White, Matern

from mgp.model.likelihoods import Gaussian

from mgp.model.options import options
from mgp.model.train import train

from mgp.model.train import evaluate
from mgp.model.utils import save_model, load_model
from mgp.model.layers import SVGPLayer
import matplotlib.pyplot as plt
from torch.nn import ModuleList
from sklearn.cluster import KMeans
import random
from mgp.model.dgp import DGP

seed_value = 0
# 1. Set the `PYTHONHASHSEED` environment variable at a fixed value
import os

np.random.seed(0)
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"


opt = options().parse()
method = opt.name.lower()

if method == "svgp":
    layer_used = SVGPLayer
else:
    layer_used = None


if opt.kernel == "matern":
    kernel_used = Matern
elif opt.kernel == "rbf":
    kernel_used = RBF


if torch.cuda.is_available() and opt.nGPU >= 0:
    device = torch.device("cuda:{}".format(opt.nGPU))
    torch.cuda.set_device(opt.nGPU)
    torch.backends.cudnn.benchmark = True
else:
    device = torch.device("cpu")

print("devcie :{}".format(device.type))

initial_seed = True
if initial_seed:
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
    seed_value = 0
    os.environ["PYTHONHASHSEED"] = str(seed_value)
    # os.environ["CUBLAS_WORKSPACE_CONFIG"] = ':4096:8'
    seed = 0

    random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
        torch.cuda.manual_seed(seed)
        # torch.backends.cudnn.benchmark = False

    np.random.seed(seed)
    # torch.use_deterministic_algorithms(True)


def compute_nn(y_t, mean_t, std_t):
    f1 = np.log(std_t**2) / 2
    f2 = (y_t - mean_t) ** 2 / (2 * std_t**2)
    f3 = np.log(2 * np.pi) / 2
    # print(std_t.mean())
    return np.mean(f1 + f2 + f3)


def compute_nn_mse(y, mean_pred, std_pred, ind_nan_target):
    D = ind_nan_target.shape[1]
    nll, rmse = [], []
    for d in range(D):
        if ind_nan_target[:, d].sum() > 0:
            a = np.sqrt(
                np.mean(
                    (y[ind_nan_target[:, d], d] - mean_pred[ind_nan_target[:, d], d])
                    ** 2
                )
            )

            b = compute_nn(
                y[ind_nan_target[:, d], d],
                mean_pred[ind_nan_target[:, d], d],
                std_pred[ind_nan_target[:, d], d],
            )

            nll.append(b)
            rmse.append(a)

    nll = np.mean(nll)
    rmse = np.mean(rmse)
    # return nll[~nll.isnan()].sum(), rmse[~rmse.isnan()].sum()
    return rmse, nll


def build_model_missing(
    opt,
    X_train=None,
    y_train=None,
    y_mean=None,
    y_std=None,
    ind_total_nan=None,
    x_means=None,
    x_stds=None,
    ind_total_nan_test=None,
    ind_binary=None,
    scaler=None,
):
    empty_cluster = True
    n_rep = 0
    while empty_cluster:
        try:
            n_rep += 1
            if n_rep > 20:
                ind_sel = np.random.choice(X_train.shape[0], opt.M2)
                Z = X_train[ind_sel, :]
                break
            # Z2 = kmeans2(X_train, opt.M2, minit='points')[0]
            kmeans = KMeans(n_clusters=opt.M2, random_state=42, n_init=10).fit(X_train)
            Z = kmeans.cluster_centers_
            empty_cluster = False
        except Warning:
            pass
    if X_train.shape[0] < 1000:
        value = np.expand_dims(np.sum(np.square(X_train), 1), 1)
        distance = (
            value - 2 * np.matmul(X_train, np.transpose(X_train)) + np.transpose(value)
        )
        median = np.percentile(distance, q=50)
    else:
        median = 10

    lengthscale = np.log(median)

    # We construct the network
    D = X_train.shape[1]

    total_mis = ind_total_nan.sum(0)
    np.where(total_mis > 0)
    args = np.argsort(
        -total_mis[np.where(total_mis > 0)[0]]
    )  # sort from minimum to maximum missing
    layers_missing = np.where(total_mis > 0)[0][
        args
    ]  # np.where(ind_total_nan.sum(0)>0)[0]
    print("before", layers_missing)
    ind_std = np.argsort(x_stds[layers_missing])
    layers_missing = layers_missing[ind_std]
    print("after", layers_missing)
    num_data = []
    global layer_missing_index
    layer_missing_index = []
    num_layers_missing = layers_missing.shape[0]
    if opt.missing:
        num_data = []
        layer_sizes = []
        likelihood_gaussian = ModuleList()
        kernels = torch.nn.ModuleList()
        for layer in range(0, num_layers_missing):
            nD = D - 1
            layer_missing_index.append(layers_missing[layer])
            num_data.append((~ind_total_nan[:, layers_missing[layer]]).sum())
            # num_data.append(X_train.shape[0])

            kernels.append(
                kernel_used(
                    lengthscales=np.tile(lengthscale, nD), variance=1, device=device
                )
                + White(variance=opt.var_noise, device=device)
            )
            layer_sizes.append(nD)
            likelihood_gaussian.append(Gaussian(opt.likelihood_var, device=device))

        X_train = torch.from_numpy(X_train).to(torch.float64).to(device)
        y_train = torch.from_numpy(y_train).to(torch.float64).to(device)
        Z = torch.from_numpy(Z).to(torch.float64).to(device)

        model_missing = DGP(
            X_train,
            y_train,
            Z,
            kernels,
            layer_sizes,
            likelihood_gaussian=likelihood_gaussian,
            likelihood_ber=None,
            num_data=np.array(num_data),
            num_samples=opt.n_samples,
            y_mean=y_mean,
            y_std=y_std,
            warm_up_iters=0,
            device=device,
            ind_total_nan=ind_total_nan,
            layer_missing_index=layer_missing_index,
            ind_binary=ind_binary,
            x_means=x_means,
            x_stds=x_stds,
            scaler=scaler,
            layer=layer_used,
            use_missing_gp=True,
        )
        return model_missing


def build_model(
    opt,
    X_train=None,
    y_train=None,
    y_mean=None,
    y_std=None,
    ind_total_nan=None,
    x_means=None,
    x_stds=None,
    ind_total_nan_test=None,
    ind_binary=None,
    scaler=None,
):
    if opt.minibatch_size is None:
        opt.minibatch_size = min(1e4, X_train.shape[0])
    if opt.nolayers == 1:
        opt.n_samples = 1
    warnings.filterwarnings("error")
    if opt.fitting:
        empty_cluster = True
        n_rep = 0
        while empty_cluster:
            try:
                n_rep += 1
                if n_rep > 20:
                    ind_sel = np.random.choice(X_train.shape[0], opt.M)
                    Z = X_train[ind_sel, :]
                    break

                kmeans = KMeans(n_clusters=opt.M, random_state=0).fit(X_train)
                Z = kmeans.cluster_centers_
                empty_cluster = False
            except Warning:
                pass

    else:
        ind_sel = np.random.choice(X_train.shape[0], opt.M)
        Z = X_train[ind_sel, :]

    warnings.resetwarnings()

    if X_train.shape[0] < 1000:
        value = np.expand_dims(np.sum(np.square(X_train), 1), 1)
        distance = (
            value - 2 * np.matmul(X_train, np.transpose(X_train)) + np.transpose(value)
        )
        median = np.percentile(distance, q=50)
    else:
        median = 10

    lengthscale = np.log(median)

    # We construct the network
    D = X_train.shape[1]

    total_mis = ind_total_nan.sum(0)
    np.where(total_mis > 0)
    args = np.argsort(
        -total_mis[np.where(total_mis > 0)[0]]
    )  # sort form minimum missing to maximum missing
    layers_missing = np.where(total_mis > 0)[0][
        args
    ]  # np.where(ind_total_nan.sum(0)>0)[0]
    ind_std = np.argsort(x_stds[layers_missing])
    layers_missing = np.arange(X_train.shape[1])
    num_data = []
    global layer_missing_index

    layer_missing_index = []
    num_layers_missing = 0

    if method == "nn":
        layer_used = SVGP_NN
    elif method == "svgp":
        layer_used = SVGPLayer
    else:
        layer_used = None

    model_missing = None
    if method != "nn":
        num_data = []
        layer_sizes = []
        likelihood_gaussian = ModuleList()
        kernels = torch.nn.ModuleList()
        for layer in range(0, opt.nolayers):
            nD = D

            layer_missing_index.append(layers_missing[layer])
            # num_data.append((~ind_total_nan[:, layers_missing[layer]]).sum())
            # num_data.append(X_train.shape[0])

            kernels.append(
                kernel_used(
                    lengthscales=np.tile(lengthscale, nD), variance=100, device=device
                )
                + White(variance=opt.var_noise, device=device)
            )
            layer_sizes.append(nD)
        num_data = [X_train.shape[0]]
        # for layer in range(num_layers_missing):
        #   if layer < num_layers_missing:
        #      layer_missing_index.append(layers_missing[layer])
        #     num_data.append((~ind_total_nan[:, layers_missing[layer]]).sum())
        likelihood_gaussian.append(Gaussian(opt.likelihood_var, device=device))
        X_train = torch.from_numpy(X_train).to(torch.float64).to(device)
        y_train = torch.from_numpy(y_train).to(torch.float64).to(device)
        Z = torch.from_numpy(Z).to(torch.float64).to(device)
        # options().parse()
        model = DGP(
            X_train,
            y_train,
            Z,
            kernels,
            layer_sizes,
            likelihood_gaussian=likelihood_gaussian,
            likelihood_ber=None,
            num_data=np.array(num_data),
            num_samples=opt.n_samples,
            y_mean=y_mean,
            y_std=y_std,
            warm_up_iters=0,
            device=device,
            ind_total_nan=ind_total_nan,
            layer_missing_index=layer_missing_index,
            ind_binary=ind_binary,
            x_means=x_means,
            x_stds=x_stds,
            scaler=scaler,
            layer=layer_used,
            use_missing_gp=False,

        )

    else:
        raise NotImplementedError

    return model


def launch_experiment(opt):
    likelihoods, rmses, times = [], [], []

    [
        X_train,
        X_test,
        ind_total_nan,
        imp,
        ind_binary,
        x_means,
        x_stds,
        ind_total_nan_test,
        scaler,
        y_train,
        y_test,
        y_mean,
        y_std,
        X_train_full,
        X_test_full,
    ] = load_dataset(opt)
    if opt.missing:
        add = "missing"
    else:
        add = "non_missing"

    if opt.missing:
        training_generator_missing = customLoader(
            X_train,
            X_train_full,
            ind_total_nan,
            ind_total_nan,
            opt.minibatch_size,
            numThreads=opt.numThreads,
            shuffle=True,
            device=device,
        )

        test_generator_missing = customLoader(
            X_test,
            X_test_full,
            ind_total_nan_test,
            ind_total_nan_test,
            opt.minibatch_size,
            numThreads=opt.numThreads,
            shuffle=False,
            device=device,
        )

    # test_generator = customLoader(X_test, None, ind_total_nan_test, opt.minibatch_size,
    #                             numThreads=opt.numThreads, shuffle=False, device=device)
    path_results = "results/{}/{}/s{}_{}_{}_{}/".format(
        opt.name.lower(),
        opt.dataset_name,
        opt.split_number,
        add,
        opt.imputation,
        opt.nolayers,
    )

    N = X_train.shape[0]
    if not os.path.exists(path_results):
        os.makedirs(path_results)

    file_name = "results.txt"

    global net
    global net_missing
    if opt.missing:
        net_missing = build_model_missing(
            opt,
            X_train=X_train,
            y_train=X_train,
            y_mean=torch.from_numpy(np.array(x_means)).to(device).to(torch.float64),
            y_std=torch.from_numpy(np.array(x_stds)).to(device).to(torch.float64),
            ind_total_nan=ind_total_nan,
            x_means=x_means,
            x_stds=x_stds,
            ind_total_nan_test=ind_total_nan_test,
            ind_binary=ind_binary,
            scaler=scaler,
        )

        params = list(net_missing.parameters())
        print(
            "There are {} trainable parameters for missing value".format(
                np.sum([par.numel() for par in params])
            )
        )
        optimizer = torch.optim.Adam(
            net_missing.parameters(), lr=opt.lrate, weight_decay=0.00
        )
        path_file_name = os.path.join(path_results, file_name)
        n_data_points = X_train.shape[0]
        iter_per_epoch = n_data_points / opt.minibatch_size
        epoch = int(np.ceil(opt.no_iterations / iter_per_epoch))

        X_train = np.tile(np.expand_dims(X_train, 0), [opt.n_samples, 1, 1])
        X_test = np.tile(np.expand_dims(X_test, 0), [opt.n_samples, 1, 1])

        training_generator_missing = customLoader(
            X_train,
            X_train_full,
            ind_total_nan,
            ind_total_nan,
            opt.minibatch_size,
            numThreads=opt.numThreads,
            shuffle=True,
            device=device,
        )

        test_generator_missing = customLoader(
            X_test,
            X_test_full,
            ind_total_nan_test,
            ind_total_nan_test,
            opt.minibatch_size,
            numThreads=opt.numThreads,
            shuffle=False,
            device=device,
        )
        if opt.fitting:
            loggs = train(
                net_missing,
                training_generator_missing,
                test_generator_missing,
                optimizer,
                path_file_name,
                opt.minibatch_size,
                epochs=epoch,
                device=device,
                layer_missing_index=layer_missing_index,
                lrate=opt.lrate,
                ind_binary=ind_binary,
                predict_test=False,
                verbose=True
            )
        else:
            load_model(
                net_missing,
                optimizer,
                path_results,
                which_epoch="latest_missing",
                device=device,
            )
            loggs = train(
                net_missing,
                training_generator_missing,
                test_generator_missing,
                optimizer,
                path_file_name,
                opt.minibatch_size,
                epochs=0,
                device=device,
                layer_missing_index=layer_missing_index,
                lrate=opt.lrate,
                ind_binary=ind_binary,
                predict_test=False,
            )
        training_generator_missing = customLoader(
            X_train,
            X_train_full,
            ind_total_nan,
            ind_total_nan,
            opt.minibatch_size,
            numThreads=opt.numThreads,
            shuffle=True,
            device=device,
        )

        test_generator_missing = customLoader(
            X_test,
            X_test_full,
            ind_total_nan_test,
            ind_total_nan_test,
            opt.minibatch_size,
            numThreads=opt.numThreads,
            shuffle=False,
            device=device,
        )
        (
            NELBO_train,
            NLL_train,
            ACC_train,
            NELBO_test,
            NLL_test,
            ACC_test,
            train_info,
            test_info,
        ) = evaluate(
            net_missing,
            training_generator_missing,
            test_generator_missing,
            path_file_name,
            loggs,
            batch_size=opt.minibatch_size,
            device=device,
        )
        [
            x_aggr_train,
            y_aggr_train,
            mean_pred_aggr_train,
            std_pred_aggr_train,
            ind_nan_aggr_train,
        ] = train_info
        [
            x_aggr_test,
            y_aggr_test,
            mean_pred_aggr_test,
            std_pred_aggr_test,
            ind_nan_aggr_test,
        ] = test_info
        # print(y_mst_test.shape)
        X_test_0 = y_aggr_test
        # y_mst_test_0 = np.mean(y_mst_test, 0)
        rmse_test, nll_test = compute_nn_mse(
            X_test_0,
            np.mean(mean_pred_aggr_test, 0),
            np.mean(std_pred_aggr_test, 0),
            ind_nan_aggr_test,
        )
        print("new_rmse {}, new nll {}".format(rmse_test, nll_test))

        X_train = x_aggr_train * x_stds + x_means
        X_train_full = y_aggr_train * x_stds + x_means
        X_test = x_aggr_test * x_stds + x_means
        X_test_full = y_aggr_test * x_stds + x_means
        # print(y_mst_train[:,0].shape)
        ind_total_nan = ind_nan_aggr_train
        ind_total_nan_test = ind_nan_aggr_test
        s, n, d = mean_pred_aggr_train.shape
        y_mst_train_conv = (
            mean_pred_aggr_train + np.random.randn(s, n, d) * std_pred_aggr_train
        )
        X_train[:, ind_total_nan] = y_mst_train_conv[:, ind_total_nan]

        s, n, d = mean_pred_aggr_test.shape
        y_mst_test_conv = (
            mean_pred_aggr_test + np.random.randn(s, n, d) * std_pred_aggr_test
        )
        X_test[:, ind_total_nan_test] = y_mst_test_conv[:, ind_total_nan_test]

        change_mean_std = False
        if change_mean_std:
            X_t = np.mean(X_train, 0)
            x_means = np.mean(X_t, 0)
            x_stds = X_t.std(0)
            net_missing.x_means = x_means
            net_missing.y_mean = (
                torch.from_numpy(np.array(x_means)).to(device).to(torch.float64)
            )
            net_missing.y_std = (
                torch.from_numpy(np.array(x_stds)).to(device).to(torch.float64)
            )
            net_missing.x_stds = x_stds

        # print(x_stds)
        X_train = (X_train - x_means) / x_stds
        X_test = (X_test - x_means) / x_stds
        X_train_full = (X_train_full - x_means) / x_stds
        X_test_full = (X_test_full - x_means) / x_stds
    else:
        net = build_model(
            opt,
            X_train=X_train,
            y_train=X_train,
            y_mean=torch.from_numpy(np.array(x_means)).to(device).to(torch.float64),
            y_std=torch.from_numpy(np.array(x_stds)).to(device).to(torch.float64),
            ind_total_nan=ind_total_nan,
            x_means=x_means,
            x_stds=x_stds,
            ind_total_nan_test=ind_total_nan_test,
            ind_binary=ind_binary,
            scaler=scaler,
        )
        params = list(net.parameters())
        print(
            "There are {} trainable parameters for missing value".format(
                np.sum([par.numel() for par in params])
            )
        )
        optimizer = torch.optim.Adam(net.parameters(), lr=opt.lrate, weight_decay=0.00)
        path_file_name = os.path.join(path_results, file_name)
        n_data_points = X_train.shape[0]
        iter_per_epoch = n_data_points / opt.minibatch_size
        epoch = int(np.ceil(opt.no_iterations / iter_per_epoch))

        X_train = np.tile(np.expand_dims(X_train, 0), [opt.n_samples, 1, 1])
        X_test = np.tile(np.expand_dims(X_test, 0), [opt.n_samples, 1, 1])
        training_generator_missing = customLoader(
            X_train,
            X_train_full,
            ind_total_nan,
            ind_total_nan,
            opt.minibatch_size,
            numThreads=opt.numThreads,
            shuffle=True,
            device=device,
        )

        test_generator_missing = customLoader(
            X_test,
            X_test_full,
            ind_total_nan_test,
            ind_total_nan_test,
            opt.minibatch_size,
            numThreads=opt.numThreads,
            shuffle=False,
            device=device,
        )
        if opt.fitting:
            loggs = train(
                net,
                training_generator_missing,
                test_generator_missing,
                optimizer,
                path_file_name,
                opt.minibatch_size,
                epochs=epoch,
                device=device,
                layer_missing_index=layer_missing_index,
                lrate=opt.lrate,
                ind_binary=ind_binary,
                predict_test=False,
                verbose=True
            )
        else:
            load_model(
                net, optimizer, path_results, which_epoch="latest", device=device
            )
            loggs = train(
                net,
                training_generator_missing,
                test_generator_missing,
                optimizer,
                path_file_name,
                opt.minibatch_size,
                epochs=0,
                device=device,
                layer_missing_index=layer_missing_index,
                lrate=opt.lrate,
                ind_binary=ind_binary,
                predict_test=False,
            )

        training_generator_missing = customLoader(
            X_train,
            X_train_full,
            ind_total_nan,
            ind_total_nan,
            opt.minibatch_size,
            numThreads=opt.numThreads,
            shuffle=True,
            device=device,
        )

        test_generator_missing = customLoader(
            X_test,
            X_test_full,
            ind_total_nan_test,
            ind_total_nan_test,
            opt.minibatch_size,
            numThreads=opt.numThreads,
            shuffle=False,
            device=device,
        )
        (
            NELBO_train,
            NLL_train,
            ACC_train,
            NELBO_test,
            NLL_test,
            ACC_test,
            train_info,
            test_info,
        ) = evaluate(
            net,
            training_generator_missing,
            test_generator_missing,
            path_file_name,
            loggs,
            batch_size=opt.minibatch_size,
            device=device,
        )
        [
            x_aggr_train,
            y_aggr_train,
            mean_pred_aggr_train,
            std_pred_aggr_train,
            ind_nan_aggr_train,
        ] = train_info
        [
            x_aggr_test,
            y_aggr_test,
            mean_pred_aggr_test,
            std_pred_aggr_test,
            ind_nan_aggr_test,
        ] = test_info
        print(ind_total_nan_test.shape)
        # y_mst_test_0 = np.mean(y_mst_test, 0)
        X_test_0 = y_aggr_test
        rmse_test, nll_test = compute_nn_mse(
            X_test_0,
            np.mean(mean_pred_aggr_test, 0),
            np.mean(std_pred_aggr_test, 0),
            ind_nan_aggr_test,
        )
        print("new_rmse {}, new nll {}".format(rmse_test, nll_test))


launch_experiment(opt)



import argparse
import os


class options:
    def __init__(self):
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
        parser = self.initialize(parser)
        self.parser = parser

    def initialize(self, parser):
        parser.add_argument("--dataset_name", default=None, required=True)
        parser.add_argument("--name", default="nn", help="DSVI_Latent_Inverse | DSVI")

        parser.add_argument("--nolayers", type=int, default=0, help="number of players")

        parser.add_argument(
            "--n_samples", default=20, type=int, help="number of samples"
        )

        parser.add_argument(
            "--M", type=int, default=10, help="number of inducing points"
        )
        parser.add_argument(
            "--M2", type=int, default=10, help="number of inducing points missing"
        )
        parser.add_argument(
            "--scaling",
            type=str,
            default="MeanStd",
            help="scaling method [MeanStd|MinMax|MaxAbs|Robust|None]",
        )

        parser.add_argument(
            "--imputation",
            type=str,
            default="median",
            help="imputation method [medina|meann|knn|mice|None]",
        )

        parser.add_argument(
            "--Ptype",
            type=str,
            default="reg",
            help="Problem type (regression or classification) [reg|class]",
        )

        parser.add_argument(
            "--kernel", type=str, default="matern", help="matern|rbf|nsrbf"
        )

        parser.add_argument(
            "--no_iterations", type=int, default=1000, help="number of inducing points"
        )

        parser.add_argument(
            "--var_noise", type=float, default=1e-4, help="variance noise"
        )

        parser.add_argument("--lrate", type=float, default=1e-2, help="learning rate")

        parser.add_argument(
            "--likelihood_var",
            type=float,
            default=0.01,
            help="variance noise gaussian likelihood",
        )

        parser.add_argument(
            "--minibatch_size", type=int, default=100, help="mini-batch size"
        )

        parser.add_argument("--n_epoch", type=int, default=100, help="mini-batch size")

        parser.add_argument(
            "--numThreads", type=int, default=1, help="number of threads"
        )

        parser.add_argument(
            "--fitting", action="store_true", help="fitting the model to the data"
        )

        parser.add_argument(
            "--missing", action="store_true", help="shared parameters in nueral network"
        )

        parser.add_argument(
            "--consider_miss",
            action="store_true",
            help="shared parameters in nueral network",
        )

        parser.add_argument("--split_number", type=int, default=0, help="split number")

        parser.add_argument("--nGPU", type=int, default=0, help="GPU number")

        return parser

    def parse(self, write_conf=True):
        opt = self.parser.parse_args()

        if opt.missing:
            add = "missing"
        else:
            add = "non_missing"
        if write_conf:
            path_config = "results/{}/{}/s{}_{}_{}_{}/".format(
                opt.name.lower(),
                opt.dataset_name,
                opt.split_number,
                add,
                opt.imputation,
                opt.nolayers,
            )
            # path_config = "results/{}/{}/s{}/".format(opt.dataset_name, opt.name.upper(), opt.split_number)
            if not os.path.exists(path_config):
                os.makedirs(path_config)

            message = ""
            message += "----------------- Options ---------------\n"
            for k, v in sorted(vars(opt).items()):
                default = self.parser.get_default(k)
                comment = "\t[default: {}] \t[help: {}]\t ".format(
                    str(default), str(self.parser._option_string_actions["--" + k].help)
                )
                message += "{:>25}: {:<30}{}\n".format(str(k), str(v), comment)
            message += "----------------- End -------------------"
            print(message)
            with open(path_config + "conf.txt", "w") as file:
                file.write(message)
        self.opt = opt
        return opt



# Credit to GPFlow.

import torch
import torch.nn as nn
import numpy as np


class Gaussian(nn.Module):
    """"""

    def __init__(self, variance=1.0, dtype=torch.float64, device=None, **kwargs):
        """"""
        super().__init__()
        self.dtype = dtype
        self.device = device

        self.register_parameter(
            "log_variance",
            nn.Parameter(
                torch.tensor(np.log(variance), dtype=self.dtype, device=self.device)
            ),
        )

        self.rmse_metric = torch.nn.MSELoss()

        self.nll_metric = torch.mean

    @property
    def metrics(self):
        return [self.rmse_metric, self.nll_metric]

    def update_metrics(
        self,
        y,
        mean_pred,
        std_pred,
        ind_nan=None,
        layer_missing_index=None,
        max_use=None,
    ):
        """
        @param y: output
        @param mean_pred: mean prediction results
        @param std_pred: standard deviation prediction
        @param ind_nan: index of missing values
        @param layer_missing_index: missing index in each layer
        @return: updated metrics
        """

        def compute_nn(y_t, mean_t, std_t):
            S = mean_t.shape[0]
            normal = torch.distributions.Normal(loc=mean_t, scale=std_t)
            logpdf = normal.log_prob(y_t)
            nll = torch.logsumexp(logpdf, 0) - np.log(S)
            return -nll.mean()

        # tiled_y = y.unsqueeze(0).tile([mean_pred[0].shape[0],1,1])
        rmse_test = []
        rmse_train = []
        nll_val = []
        nll_val_test = []
        for o, el in enumerate(layer_missing_index):
            ind_nan_dim = ind_nan[:, el]
            if ind_nan_dim.sum() > 0:
                mean_test = mean_pred[o][:, ind_nan_dim, :].mean(0).squeeze(-1)
                std_test = std_pred[o][:, ind_nan_dim, :].mean(0)
                y_true_test = y[ind_nan_dim, el]

                rmse_test.append(self.rmse_metric(y_true_test, mean_test).sqrt())
                nll_val_test.append(compute_nn(y_true_test, mean_test, std_test))

                mean_train = mean_pred[o][:, ~ind_nan_dim, :].mean(0).squeeze(-1)
                std_train = std_pred[o][:, ~ind_nan_dim, :].mean(0)
                y_true_train = y[~ind_nan_dim, el]

                rmse_train.append(self.rmse_metric(y_true_train, mean_train).sqrt())
                nll_val.append(compute_nn(y_true_train, mean_train, std_train))

        self.rmse_val = torch.stack(rmse_train).mean()
        self.rmse_val_test = torch.stack(rmse_test).mean()
        self.nll_val_test = torch.stack(nll_val_test).mean()
        self.nll_val = torch.stack(nll_val).mean()

    def logdensity(self, x, mu, var):
        """
        @param x: observed value
        @param mu: prediction
        @param var: prediction variance
        @return: log density normal distribution
        """
        return -0.5 * (np.log(2 * np.pi) + var.log() + (mu - x).square() / var)

    def logp(self, F, Y):
        return self.logdensity(Y, F, self.log_variance.exp())

    def predict_mean_and_var(self, Fmu, Fvar):
        return Fmu, Fvar + self.log_variance.exp()

    def predict_density(self, Fmu, Fvar, Y):
        return self.logdensity(Y, Fmu, Fvar + self.log_variance.exp())

    def variational_expectations(self, Fmu, Fvar, Y):
        """
        @param Fmu: mean prediction
        @param Fvar: variance prediciton
        @param Y: observed values
        @return: variational expection Gaussian likelihood
        """
        return (
            -0.5 * np.log(2 * np.pi)
            - 0.5 * self.log_variance
            - (0.5 / (self.log_variance.exp()))
            * (Y.pow(2) - 2 * Y * Fmu + (Fmu.pow(2) + Fvar))
        ).sum(-1)
        # return -0.5 * np.log(2 * np.pi) - 0.5 * self.log_variance \
        #      - 0.5 * ((Y - Fmu).square() + Fvar) / self.log_variance.exp()



import torch
import torch.nn as nn


class MeanFunction(nn.Module):
    """
    Base mean function
    """

    def __init__(self, dtype, device):
        super().__init__()
        self.dtype = dtype
        self.device = device

    def __call__(self, X, ind_nan=None):
        raise NotImplementedError("Not Implemented")


class Identity(MeanFunction):
    """
    y = x
    """

    def __init__(self, input_dim=None, dtype=torch.float64, device=None):
        MeanFunction.__init__(self, dtype=dtype, device=device)
        self.input_dim = input_dim

    def __call__(self, X, ind_nan=None):
        """
        @param X: input matrix
        @param ind_nan:
        @return: the same matrix
        """
        return X


class Zero(MeanFunction):
    def __init__(self, output_dim=1, dtype=torch.float64, device=None):
        MeanFunction.__init__(self, dtype=dtype, device=device)
        self.output_dim = output_dim

    def __call__(self, X, ind_nan=None):
        """
        @param X: input
        @param ind_nan:
        @return: zeros similar to x
        """
        return torch.zeros([X.shape[0], self.output_dim]).to(X.dtype).to(self.device)



import os.path
import random
import torch
import numpy as np
import abc
import torch.utils.data as data


class baseData(data.Dataset):
    def __init__(self):
        super(baseData, self).__init__()

    @abc.abstractmethod
    def create(self, opt, x, y):
        raise NotImplementedError("Subclass should be implemented.")

    @abc.abstractmethod
    def __len__(self):
        raise NotImplementedError("Subclass should be implemented.")

    @abc.abstractmethod
    def name(self):
        raise NotImplementedError("Subclass should be implemented.")


class dataloader(baseData):
    def name(self):
        return "DataLoader"

    def create(self, X, y, ind_nan, ind_nan_target, device=None, dtype=torch.float64):
        """
        options of the data set
        :param opt:
        :return:
        """

        self.X = X
        self.y = None
        if y is not None:
            self.y = y
        self.ind_nan = ind_nan
        self.device = device
        self.dtype = dtype
        if ind_nan_target.ndim == 1:
            ind_nan_target = ind_nan_target.reshape(-1, 1)
        self.ind_nan_target = ind_nan_target

    def __getitem__(self, item):
        """
        randomly select one image form the pool
        :param item:
        :return:
        """
        X = self.X[:, item, :]

        ind_nan = self.ind_nan[item, :]
        x = torch.from_numpy(X).to(self.dtype)  # .to(self.device)

        y = self.y[item, :]
        y = torch.from_numpy(y).to(self.dtype)  # .to(self.device)
        # if y.ndim == 1:
        #   y = y.unsqueeze(-1)
        ind_nan_target = self.ind_nan_target[item, :]
        return x, y, ind_nan, ind_nan_target

    def __len__(self):
        return self.y.shape[0]



# Credit to GPflow

import torch
import torch.nn as nn
import numpy as np
from torch.nn.functional import softmax
import abc
from functools import partial, reduce
from typing import List, Optional, Union

ActiveDims = Union[slice, list]


def square_distance(X, X2):
    """
    RBF KERNEL
    """

    Xs = X.square().sum(-1).unsqueeze(-1)
    if X2 is None:
        dist = -2 * X.matmul(X.transpose(-1, -2))

        dist += Xs + Xs.transpose(-1, -2)
        return dist

    X2s = X2.square().sum(-1).unsqueeze(-1)
    dist = -2 * X.matmul(X2.transpose(-1, -2))
    dist += Xs + X2s.transpose(-1, -2)
    return dist


class Kernel(nn.Module):
    """
    base class for kernels
    """

    def name(self):
        "Kernel"

    def __init__(
        self,
        active_dims: Optional[ActiveDims] = None,
        name: Optional[str] = None,
        dtype=torch.float64,
        device=None,
    ):
        """ """
        super().__init__()
        self.device = device
        self.dtype = dtype
        self._active_dims = self._normalize_active_dims(active_dims)

    @staticmethod
    def _normalize_active_dims(value):
        if value is None:
            value = slice(None, None, None)
        if not isinstance(value, slice):
            value = np.array(value, dtype=int)
        return value

    @property
    def active_dims(self):
        return self._active_dims

    @active_dims.setter
    def active_dims(self, value):
        self._active_dims = self._normalize_active_dims(value)

    def slice(self, X: torch.Tensor, X2: Optional[torch.Tensor] = None):
        """ """
        dims = self.active_dims
        if isinstance(dims, slice):
            X = X[..., dims]
            if X2 is not None:
                X2 = X2[..., dims]

        return X, X2

    def _validate_ard_active_dims(self, ard_parameter):
        """
        Validate that ARD parameter matches the number of active_dims (provided active_dims
        has been specified as an array).
        """
        if self.active_dims is None or isinstance(self.active_dims, slice):
            # Can only validate parameter if active_dims is an array
            return

        if ard_parameter.shape.rank > 0 and ard_parameter.shape[0] != len(
            self.active_dims
        ):
            raise ValueError(
                f"Size of `active_dims` {self.active_dims} does not match "
                f"size of ard parameter ({ard_parameter.shape[0]})"
            )

    @abc.abstractmethod
    def K(self, X, X2=None):
        raise NotImplementedError

    @abc.abstractmethod
    def K_diag(self, X):
        raise NotImplementedError

    def __call__(self, X, X2=None, *, full_cov=True, presliced=False):
        if (not full_cov) and (X2 is not None):
            raise ValueError(
                "Ambiguous inputs: `not full_cov` and `X2` are not compatible."
            )

        if not presliced:
            X, X2 = self.slice(X, X2)

        if not full_cov:
            assert X2 is None
            return self.K_diag(X)

        else:
            return self.K(X, X2)

    def __add__(self, other):
        return Sum([self, other])


class Combination(Kernel):
    """ """

    _reduction = None

    def __init__(self, kernels: List[Kernel], name: Optional[str] = None):
        super().__init__(name=name)

        if not all(isinstance(k, Kernel) for k in kernels):
            raise TypeError("can only combine Kernel instances")  # pragma: no cover

        self._set_kernels(kernels)

    def _set_kernels(self, kernels: List[Kernel]):
        kernels_list = nn.ModuleList()
        for k in kernels:
            if isinstance(k, self.__class__):
                kernels_list.extend(k.kernels)
            else:
                kernels_list.append(k)
        self.kernels = kernels_list

    @property
    def on_separate_dimensions(self):
        """ """
        if np.any([isinstance(k.active_dims, slice) for k in self.kernels]):
            return False
        else:
            dimlist = [k.active_dims for k in self.kernels]
            overlapping = False
            for i, dims_i in enumerate(dimlist):
                for dims_j in dimlist[i + 1 :]:
                    print(f"dims_i = {type(dims_i)}")
                    if np.any(dims_i.reshape(-1, 1) == dims_j.reshape(1, -1)):
                        overlapping = True
            return not overlapping


class ReducingCombination(Combination):
    def __call__(self, X, X2=None, *, full_cov=True, presliced=False):
        return self._reduce(
            [k(X, X2, full_cov=full_cov, presliced=presliced) for k in self.kernels]
        )

    def K(self, X: torch.Tensor, X2: Optional[torch.Tensor] = None) -> torch.Tensor:
        return self._reduce([k.K(X, X2) for k in self.kernels])

    def K_diag(self, X: torch.Tensor) -> torch.Tensor:
        return self._reduce([k.K_diag(X) for k in self.kernels])

    @property
    @abc.abstractmethod
    def _reduce(self):
        pass


class Sum(ReducingCombination):
    @property
    def _reduce(self):
        return sum


class Stationary(Kernel):
    """ """

    def __init__(self, variance=1.0, lengthscales=1.0, device=None, **kwargs):
        """ """
        for kwarg in kwargs:
            if kwarg not in {"name", "active_dims"}:
                raise TypeError(f"Unknown keyword argument: {kwarg}")

        super().__init__(device=device, **kwargs)

        self.register_parameter(
            "log_variance",
            nn.Parameter(
                torch.tensor(np.log(variance), dtype=self.dtype, device=self.device)
            ),
        )

        self.register_parameter(
            "log_lengthscales",
            nn.Parameter(
                torch.tensor(np.log(lengthscales), dtype=self.dtype, device=self.device)
            ),
        )

        self._validate_ard_active_dims(self.log_lengthscales)

    @property
    def ard(self) -> bool:
        """ """
        return self.log_lengthscales.shape.ndims > 0

    def scale(self, X):
        X_scaled = X / torch.exp(self.log_lengthscales) if X is not None else X
        return X_scaled

    def K_diag(self, X):
        return (self.log_variance).exp().repeat(X.shape[-2])


class IsotropicStationary(Stationary):
    """ """

    def K(self, X, X2=None):
        r2 = self.scaled_squared_euclid_dist(X, X2)
        return self.K_r2(r2)

    def scaled_squared_euclid_dist(self, X, X2=None):
        """ """
        return square_distance(self.scale(X), self.scale(X2))


class Static(Kernel):
    """ """

    def __init__(self, variance=1.0, active_dims=None, device=None):
        super().__init__(active_dims, device=device)

        self.register_parameter(
            "log_variance",
            nn.Parameter(
                torch.tensor(np.log(variance), dtype=self.dtype, device=self.device)
            ),
        )

    def K_diag(self, X):
        return (self.log_variance).exp().repeat(X.shape[-2])


class White(Static):
    """ """

    def K(self, X, X2=None):
        if X2 is None:
            d = torch.tile((self.log_variance).exp().squeeze(), dims=(X.shape[0], 1))[
                :, 0
            ]
            return torch.diag_embed(d)
        else:
            return torch.zeros((X.shape[-2], X2.shape[-2])).to(X.device).to(X.dtype)


class SquaredExponential(IsotropicStationary):
    """ """

    def K_r2(self, r2):
        return (self.log_variance).exp() * torch.exp(-0.5 * r2)


class Matern(IsotropicStationary):
    """
    @
    """

    def K(self, X, X2=None):
        r = self.scaled_euclid_dist(X, X2)
        return self.K_r(r)

    def scaled_euclid_dist(self, X, X2=None):
        """ """
        if X2 is None:
            return torch.cdist(self.scale(X), self.scale(X), p=2).sqrt()
        return torch.cdist(self.scale(X), self.scale(X2), p=2).sqrt()

    def K_r(self, r, nu=1.5):
        # multiply by term
        exp_c = torch.exp(-np.sqrt(nu * 2) * r)

        if nu == 0.5:
            # Matern 1/2
            const_c = 1
        elif nu == 1.5:
            # Matern 3/2
            const_c = 1.0 + np.sqrt(3) * r
        elif nu == 2.5:
            # Matern 5/2
            const_c = np.sqrt(5) * r + 1.0 + 5.0 / 3.0 * r**2

        return (self.log_variance).exp() * const_c * exp_c



import numpy as np

import torch.nn as nn
from mgp.model.mean_functions import Identity, Zero
from mgp.model.layers import SVGPLayer


def init_layers_linear(
    X,
    Y,
    Z,
    kernels,
    layer_sizes,
    mean_function=None,
    num_outputs=None,
    Layer=SVGPLayer,
    whiten=True,
    num_latent=None,
    num_samples=None,
    shared_nn=False,
    device=None,
    ind_total_nan=None,
    layer_missing_index=None,
    opt=None,
):
    """
    initialization of the DGP layers
    @param X:
    @param Y:
    @param Z:
    @param kernels:
    @param layer_sizes:
    @param mean_function:
    @param num_outputs:
    @param Layer:
    @param whiten:
    @param num_latent:
    @param num_samples:
    @param shared_nn:
    @param device:
    @param ind_total_nan:
    @param layer_missing_index:
    @param opt:
    @return:
    """

    kernels_used = kernels

    layers = nn.ModuleList()
    X_running, Z_running = X.clone().detach(), Z.clone().detach().detach()
    if mean_function is None:
        mean_function = Zero(device=device).to(device)

    in_idx = 0
    for kern_in in kernels_used:
        dim_in = layer_sizes[in_idx]
        dim_out = dim_in

        Z_used = Z_running
        X_used = X_running

        # Initialize mean function to be either Identity or PCA projection
        if dim_in == dim_out and in_idx != len(kernels) - 1:
            mf = Identity(device=device).to(device)
        else:
            mf = Zero(device=device).to(device)
        layer = Layer(kern_in, Z_used, dim_out, mf, white=whiten, device=device)
        layers.append(layer)
        in_idx += 1

    return layers


def init_layers_linear_missing(
    X,
    Y,
    Z,
    kernels,
    layer_sizes,
    mean_function=None,
    num_outputs=None,
    Layer=SVGPLayer,
    whiten=True,
    num_latent=None,
    num_samples=None,
    shared_nn=False,
    device=None,
    ind_total_nan=None,
    layer_missing_index=None,
    opt=None,
):
    """
    initialization of the DGP layers for MGP
    @param X:
    @param Y:
    @param Z:
    @param kernels:
    @param layer_sizes:
    @param mean_function:
    @param num_outputs:
    @param Layer:
    @param whiten:
    @param num_latent:
    @param num_samples:
    @param shared_nn:
    @param device:
    @param ind_total_nan:
    @param layer_missing_index:
    @param opt:
    @return:
    """
    kernels_used = kernels
    total_dim = X.shape[1]
    layers = nn.ModuleList()
    X_running, Z_running = X.clone().detach(), Z.clone().detach().detach()
    in_idx = 0
    for kern_in in kernels_used:
        layer_miss_ind = layer_missing_index[in_idx]
        dim_out = 1  # layer_sizes[in_idx + 1]
        dim_in = layer_sizes[in_idx]

        Z_used = Z_running
        mf = Zero(device=device).to(device)
        other_inds = np.array((list(set(np.arange(total_dim)) - set({layer_miss_ind}))))
        layer = Layer(
            kern_in, Z_used[:, other_inds], dim_out, mf, white=whiten, device=device
        )

        layers.append(layer)
        in_idx += 1

    return layers



import torch
import torch.nn as nn

from mgp.model.layer_initializations import (
    init_layers_linear,
    init_layers_linear_missing,
)
import numpy as np


class DGP_Base(nn.Module):
    def name(self):
        return "DGP_Base"

    def __init__(
        self,
        likelihood_gaussian,
        likelihood_ber,
        layers,
        num_data,
        num_samples=1,
        y_mean=0.0,
        y_std=1.0,
        warm_up_iters=None,
        dtype=torch.float64,
        seed=0,
        device=None,
        ind_total_nan=None,
        layer_missing_index=None,
        ind_binary=None,
        x_means=None,
        x_stds=None,
        **kwargs
    ):
        super().__init__()
        self.x_means = x_means
        self.x_stds = x_stds
        self.device = device
        self.dtype = dtype
        self.num_samples = num_samples
        self.num_data = num_data
        self.y_mean = y_mean
        self.y_std = y_std
        self.back_transform = False
        self.warm_up_iters = warm_up_iters

        self.likelihood_gaussian = likelihood_gaussian
        self.likelihood_bern = likelihood_ber
        self.layers_dgp = layers
        torch.manual_seed(seed)
        self.use_interconnection = False
        self.max_use = 1
        self.current_dim = 0
        self.ind_total_nan = ind_total_nan
        self.layer_missing_index = layer_missing_index
        self.ind_binary = ind_binary
        self.rmse_metric = torch.nn.MSELoss()
        self.no_gt = False

    def train_step(
        self,
        optimizer,
        batch_x,
        batch_y,
        ind_nan,
        ind_nan_target,
        current_iter,
        max_use,
    ):
        """
        Training step
        @param optimizer:
        @param batch_x:
        @param batch_y:
        @param ind_nan:
        @param ind_nan_target:
        @param current_iter:
        @param max_use:
        @return:
        """
        if batch_y.ndim == 1:
            batch_y = batch_y.unsqueeze(-1)

        if self.dtype != batch_x.dtype:
            batch_x = batch_x.to(self.dtype)
        if self.dtype != batch_y.dtype:
            batch_y = batch_y.to(self.dtype)

        loss = self.nelbo(
            batch_x, batch_y, ind_nan, ind_nan_target, current_iter, max_use=max_use
        )
        with torch.no_grad():
            means, std_pred = self.forward(batch_x, ind_nan=ind_nan, max_use=max_use)
        optimizer.zero_grad()
        loss.backward()  # retain_graph=True

        optimizer.step()
        if self.back_transform:
            self.update_metrics(
                batch_y * self.y_std + self.y_mean,
                means,
                std_pred,
                ind_nan=ind_nan,
                ind_nan_target=ind_nan_target,
                layer_missing_index=self.layer_missing_index,
            )
        else:
            self.update_metrics(
                batch_y,
                means,
                std_pred,
                ind_nan=ind_nan,
                ind_nan_target=ind_nan_target,
                layer_missing_index=self.layer_missing_index,
            )

        return loss

    def update_metrics(
        self,
        y,
        mean_pred,
        std_pred,
        ind_nan=None,
        ind_nan_target=None,
        layer_missing_index=None,
        max_use=None,
    ):
        """
        updating metrics
        @param y:
        @param mean_pred:
        @param std_pred:
        @param ind_nan:
        @param ind_nan_target:
        @param layer_missing_index:
        @param max_use:
        @return:
        """

        def compute_nn(y_t, mean_t, std_t):
            f1 = torch.log(std_t**2) / 2
            f2 = (y_t - mean_t) ** 2 / (2 * std_t**2)
            f3 = np.log(2 * np.pi) / 2

            return torch.mean(f1 + f2 + f3)

        def compute_nn_mse(y, mean_pred, std_pred, ind_nan_target):
            ind_true_target = ~ind_nan_target
            D = ind_nan_target.shape[1]
            nll, rmse = [], []
            for d in range(D):
                if ind_nan_target[:, d].sum() > 0:
                    a = self.rmse_metric(
                        y[ind_true_target[:, d], d],
                        mean_pred[:, ind_true_target[:, d], d].mean(0),
                    ).sqrt()
                    b = compute_nn(
                        y[ind_true_target[:, d], d],
                        mean_pred[:, ind_true_target[:, d], d].squeeze(-1),
                        std_pred[:, ind_true_target[:, d], d].squeeze(-1),
                    )

                    nll.append(b)
                    rmse.append(a)

            nll = torch.stack(nll)
            rmse = torch.stack(rmse)

            return nll.mean(), rmse.mean()

        if not self.onelayer_out:
            self.mean_pred_aggr = torch.stack(mean_pred, dim=2).squeeze(-1).mean(0)
            self.std_pred_aggr = torch.stack(std_pred, dim=2).squeeze(-1).mean(0)
            ind_nan_aggr = []
            y_aggr = []
            for o, el in enumerate(layer_missing_index):
                ind_nan_aggr.append(ind_nan[:, el])
                y_aggr.append(y[:, el])
            self.ind_nan_aggr = torch.stack(ind_nan_aggr, dim=1)
            self.y_aggr = torch.stack(y_aggr, dim=1)

            self.rmse_val = self.rmse_metric(
                self.y_aggr[~self.ind_nan_aggr], self.mean_pred_aggr[~self.ind_nan_aggr]
            ).sqrt()

            self.nll_val = compute_nn(
                self.y_aggr[~self.ind_nan_aggr],
                self.mean_pred_aggr[~self.ind_nan_aggr],
                self.std_pred_aggr[~self.ind_nan_aggr],
            )

            self.rmse_val_test = self.rmse_metric(
                self.y_aggr[self.ind_nan_aggr], self.mean_pred_aggr[self.ind_nan_aggr]
            ).sqrt()
            self.nll_val_test = compute_nn(
                self.y_aggr[self.ind_nan_aggr],
                self.mean_pred_aggr[self.ind_nan_aggr],
                self.std_pred_aggr[self.ind_nan_aggr],
            )

        else:
            self.mean_pred_aggr = mean_pred
            self.std_pred_aggr = std_pred

            if self.use_missing_gp:
                self.y_aggr = y
                ind_nan_target = ind_nan

                self.nll_val, self.rmse_val = compute_nn_mse(
                    y, mean_pred, std_pred, ind_nan_target
                )

                y_t = y * self.y_std + self.y_mean
                mean_pred_t = mean_pred * self.y_std + self.y_mean
                std_pred_t = std_pred * self.y_std
                self.nll_val_converted, self.rmse_val_converted = compute_nn_mse(
                    y_t, mean_pred_t, std_pred_t, ind_nan_target
                )

            else:
                self.y_aggr = y

                self.nll_val, self.rmse_val = compute_nn_mse(
                    y, mean_pred, std_pred, ind_nan_target
                )

                y_t = y * self.y_std + self.y_mean
                mean_pred_t = mean_pred * self.y_std + self.y_mean
                std_pred_t = std_pred * self.y_std
                self.nll_val_converted, self.rmse_val_converted = compute_nn_mse(
                    y_t, mean_pred_t, std_pred_t, ind_nan_target
                )

    def forward(
        self,
        inputs,
        imputing_missing_values=False,
        ind_nan=None,
        max_use=np.inf,
        num_samples=1000,
        training=False,
        back_transform=True,
    ):
        """
        @param inputs: input data
        """

        output_means, output_vars, Ds_predict = self.predict_y(
            inputs,
            self.num_samples,
            ind_nan=ind_nan,
            max_use=max_use,
            training=training,
        )
        if self.use_missing_gp:
            output_means_converted = inputs.clone()
            output_sqrt_converted = torch.zeros_like(output_means_converted)

            for l_gauss, el in enumerate(self.layer_missing_index):
                Fmean = output_means[l_gauss]
                Fvar = output_vars[l_gauss]
                D_predict = Ds_predict[l_gauss]

                Fvar += self.likelihood_gaussian[l_gauss].log_variance.exp()
                if self.back_transform:
                    Fsqrt_converted = Fvar.sqrt() * self.y_std[D_predict]
                    Fmean_converted = (
                        Fmean * self.y_std[D_predict] + self.y_mean[D_predict]
                    )
                else:
                    Fsqrt_converted = Fvar.sqrt()
                    Fmean_converted = Fmean

                output_means_converted[:, :, D_predict] = Fmean_converted.squeeze(-1)
                output_sqrt_converted[:, :, D_predict] = Fsqrt_converted.squeeze(-1)

        else:
            output_vars[-1] += self.likelihood_gaussian[-1].log_variance.exp()
            if self.back_transform:
                output_sqrt_converted = output_vars[-1].sqrt() * self.y_std
                output_means_converted = output_means[-1] * self.y_std + self.y_mean
            else:
                output_sqrt_converted = output_vars[-1].sqrt()
                output_means_converted = output_means[-1]

        return output_means_converted, output_sqrt_converted

    #
    def test_step(self, batch_x, batch_y, ind_nan, ind_nan_target, normalize_y=True):
        """
        testing the method
        @param batch_x:
        @param batch_y:
        @param ind_nan:
        @param ind_nan_target:
        @param normalize_y:
        @return:
        """
        if batch_y.ndim == 1:
            batch_y = batch_y.unsqueeze(-1)

        if self.dtype != batch_x.dtype:
            batch_x = batch_x.to(self.dtype)
        if self.dtype != batch_y.dtype:
            batch_y = batch_y.to(self.dtype)

        # Compute predictions
        with torch.no_grad():
            mean_pred, var_pred = self(
                batch_x, ind_nan=ind_nan, training=False
            )  # Forward pass
            if self.back_transform:
                self.update_metrics(
                    batch_y * self.y_std + self.y_mean,
                    mean_pred,
                    var_pred,
                    ind_nan=ind_nan,
                    ind_nan_target=ind_nan_target,
                    layer_missing_index=self.layer_missing_index,
                )
            else:
                self.update_metrics(
                    batch_y,
                    mean_pred,
                    var_pred,
                    ind_nan=ind_nan,
                    ind_nan_target=ind_nan_target,
                    layer_missing_index=self.layer_missing_index,
                )
            # Compute the loss
            loss = self.nelbo(
                batch_x, batch_y, ind_nan, ind_nan_target, 0, max_use=np.inf
            )

        return loss

    def propagate(
        self,
        X,
        full_cov=False,
        num_samples=1,
        zs=None,
        ind_nan=None,
        max_use=None,
        training=True,
    ):
        """
        computing the results
        @param X:
        @param full_cov:
        @param num_samples:
        @param zs:
        @param ind_nan:
        @param max_use:
        @param training:
        @return:
        """
        if self.use_missing_gp:
            F = X
        else:
            if X.ndim != 3:
                sX = torch.tile(X.unsqueeze(0), [num_samples, 1, 1])
                F = sX
            else:
                F = X.clone()

        zs = zs or [
            None,
        ] * len(self.layers_dgp)

        if max_use != np.inf:
            max_use = len(self.layer_missing_index) - 1

        size_loop = 1

        for iter in range(size_loop):
            Fs, Fmeans, Fvars = [], [], []
            l_n = 0
            Fmean, Fvar = [], []
            Ds_predict = []
            for layer, z in zip(self.layers_dgp, zs):
                if self.use_missing_gp:
                    D_predict = self.layer_missing_index[l_n]
                    effective_dims = list(set(np.arange(F.shape[2])) - set([D_predict]))

                    GG, Fmean, Fvar = layer.sample_from_conditional(
                        F[:, :, effective_dims],
                        z=z,
                        full_cov=full_cov,
                        max_use=max_use,
                        ind_binary=self.ind_binary,
                    )
                    Ds_predict.append(D_predict)
                    F[:, ind_nan[:, D_predict], D_predict] = (
                        GG[:, ind_nan[:, D_predict], :].squeeze(-1).clone()
                    )  # + F[:, :, D_predict].squeeze(-1)#

                else:
                    F, Fmean, Fvar = layer.sample_from_conditional(
                        F, z=z, full_cov=full_cov
                    )

                if iter == size_loop - 1:
                    Fs.append(F)
                    Fmeans.append(Fmean)
                    Fvars.append(Fvar)

                l_n += 1

        return Fs, Fmeans, Fvars, Ds_predict

    def predict_f(
        self,
        predict_at,
        num_samples,
        full_cov=False,
        ind_nan=None,
        max_use=None,
        training=True,
    ):
        Fs, Fmeans, Fvars, Ds_predict = self.propagate(
            predict_at,
            full_cov=full_cov,
            num_samples=num_samples,
            ind_nan=ind_nan,
            max_use=max_use,
            training=training,
        )
        return Fmeans, Fvars, Ds_predict

    def predict_all_layers(self, predict_at, num_samples, full_cov=False):
        return self.propagate(predict_at, full_cov=full_cov, num_samples=num_samples)

    def predict_y(
        self, predict_at, num_samples, ind_nan=None, max_use=None, training=False
    ):
        Fmean, Fvar, Ds_predict = self.predict_f(
            predict_at,
            num_samples=num_samples,
            full_cov=False,
            ind_nan=ind_nan,
            max_use=max_use,
            training=training,
        )

        return (
            Fmean,
            Fvar,
            Ds_predict,
        )  # self.likelihood_gaussian.predict_mean_and_var(Fmean[-1], Fvar[-1])

    def expected_data_log_likelihood(
        self,
        X,
        Y,
        ind_nan,
        ind_nan_target,
        index_used=None,
        max_use=np.inf,
        current_iter=0,
    ):
        """
        Compute expectations of the loglikelihood
        """
        F_means, F_vars, Ds_predict = self.predict_f(
            X,
            num_samples=self.num_samples,
            full_cov=False,
            ind_nan=ind_nan,
            max_use=max_use,
            training=True,
        )
        var_exp = 0

        if self.use_missing_gp:
            X_true = Y
            ind_total_true = ~ind_nan_target

            for l_gauss, el in enumerate(self.layer_missing_index):
                Fmeans = F_means[l_gauss]
                Fvars = F_vars[l_gauss]
                D_predict = Ds_predict[l_gauss]
                scale = (
                    self.num_data[l_gauss] / ind_total_true[:, D_predict].sum()
                )  # ind_total_true[:,o].sum()

                var_exp += (
                    scale
                    * self.likelihood_gaussian[l_gauss]
                    .variational_expectations(
                        Fmeans[:, ind_total_true[:, D_predict]].squeeze(-1),
                        Fvars[:, ind_total_true[:, D_predict]].squeeze(-1),
                        X_true[ind_total_true[:, D_predict], D_predict],
                    )
                    .mean(0)
                    .sum()
                )  # Shape [S, N, D]

        else:
            scale = self.num_data[-1] / X.shape[0]  # ind_total_true[:,o].sum()

            if self.no_gt:
                ind_total_true = ~ind_nan_target
                scale = self.num_data[-1] / ind_total_true.sum()
                var_exp += (
                    scale
                    * self.likelihood_gaussian[-1]
                    .variational_expectations(
                        F_means[-1][:, ind_total_true],
                        F_vars[-1][:, ind_total_true],
                        Y[ind_total_true],
                    )
                    .mean(0)
                    .sum()
                )
            else:
                var_exp += (
                    scale
                    * self.likelihood_gaussian[-1]
                    .variational_expectations(F_means[-1], F_vars[-1], Y)
                    .mean(0)
                    .sum()
                )
        return var_exp  # torch.mean(var_exp, dim=0)  # Shape [N, D]

    #
    def nelbo(
        self, inputs, outputs, ind_nan, ind_nan_target, current_iter, max_use=np.inf
    ):
        """
        Computes the negative elbo
        """
        X, Y = inputs, outputs

        likelihood = self.expected_data_log_likelihood(
            X, Y, ind_nan, ind_nan_target, max_use=max_use, current_iter=current_iter
        )

        if self.use_missing_gp:
            KL = torch.stack(
                [layer.KL() for num, layer in enumerate(self.layers_dgp)]
            ).sum()

        else:
            KL = torch.stack(
                [layer.KL() for num, layer in enumerate(self.layers_dgp)]
            ).sum()

        return -(likelihood - KL)  # + weight_regularizer

        # return - (-  KL)

    @property
    def metrics(self):
        metrics = [self.loss_tracker]
        for metric in self.likelihood_gaussian.metrics:
            metrics.append(metric)

        return metrics


class DGP(DGP_Base):
    """
    Doubly-Stochastic Deep GP, with linear/identity mean functions at each layer.

      @inproceedings{salimbeni2017doubly,
        title={Doubly Stochastic Variational Inference for Deep Gaussian Processes},
        author={Salimbeni, Hugh and Deisenroth, Marc},
        booktitle={NIPS},
        year={2017}
      }

    """

    def __init__(
        self,
        X,
        Y,
        Z,
        kernels,
        layer_sizes,
        likelihood_gaussian,
        likelihood_ber,
        num_data,
        y_mean=0.0,
        y_std=1.0,
        warm_up_iters=None,
        num_outputs=None,
        mean_function=None,
        whiten=False,
        num_samples=1,
        device=None,
        ind_total_nan=None,
        layer_missing_index=None,
        ind_binary=None,
        x_means=None,
        x_stds=None,
        scaler=None,
        layer=None,
        use_missing_gp=True,
    ):
        if use_missing_gp:
            inits = init_layers_linear_missing
        else:
            inits = init_layers_linear
        layers = inits(
            X,
            Y,
            Z,
            kernels,
            layer_sizes,
            mean_function=mean_function,
            num_outputs=num_outputs,
            whiten=whiten,
            device=device,
            ind_total_nan=ind_total_nan,
            layer_missing_index=layer_missing_index,
            Layer=layer,
        )
        self.use_missing_gp = use_missing_gp
        self.scaler = scaler
        self.onelayer_out = False
        if len(layer_sizes) - len(layer_missing_index) > 0:
            self.onelayer_out = True
        self.onelayer_out = True
        super().__init__(
            likelihood_gaussian,
            likelihood_ber,
            layers,
            num_data,
            num_samples,
            y_mean=y_mean,
            y_std=y_std,
            warm_up_iters=warm_up_iters,
            device=device,
            ind_total_nan=ind_total_nan,
            layer_missing_index=layer_missing_index,
            ind_binary=ind_binary,
            x_means=x_means,
            x_stds=x_stds,
        )



import torch
import numpy as np
import warnings
import os
import random

from torch.nn import ModuleList
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.cluster import KMeans
from sklearn.preprocessing import StandardScaler

# Assuming the following MGP modules are in the python path
# You might need to adjust the imports based on your project structure
from mgp.model.data import customLoader
from mgp.model.kernels import SquaredExponential as RBF
from mgp.model.kernels import White, Matern
from mgp.model.likelihoods import Gaussian
from mgp.model.train import train, evaluate
from mgp.model.dgp import DGP
from mgp.model.callback import logger
from mgp.model.layers import SVGPLayer
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import SimpleImputer
from sklearn.impute import IterativeImputer
from sklearn.linear_model import LinearRegression, BayesianRidge
from sklearn.preprocessing import RobustScaler
from sklearn.impute import KNNImputer

# Set a seed for reproducibility
def set_seed(seed):
    """Sets the seed for reproducibility."""
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
        torch.cuda.manual_seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)

class MGPImputer(BaseEstimator, TransformerMixin):
    """
    Missing Value Imputation using Deep Gaussian Processes.

    This class wraps the MGP implementation into a scikit-learn compatible
    estimator, allowing for easy integration into ML pipelines.

    Parameters
    ----------
    n_layers : int, default=1
        Number of layers in the Deep GP.
    kernel : str, default='matern'
        Kernel function to use ('rbf' or 'matern').
    imputation_strategy : str, default='chained'
        Strategy for imputation. Can be 'holistic' or 'chained'.
        'holistic': Builds a single multi-output DGP for all features.
        'chained': Builds a separate GP layer for each feature with missing values. (MGP)
    n_inducing_points : int, default=100
        Number of inducing points (M) for the sparse GP.
    n_samples : int, default=10
        Number of samples to draw for Monte Carlo estimation.
    learning_rate : float, default=0.01
        Learning rate for the Adam optimizer.
    n_iterations : int, default=2000
        Number of training iterations.
    batch_size : int, default=128
        Size of minibatches for training.
    var_noise : float, default=0.1
        Initial variance for the White kernel (noise).
    likelihood_var : float, default=0.1
        Initial variance for the Gaussian likelihood.
    use_cuda : bool, default=True
        Whether to use GPU if available.
    seed : int, default=0
        Random seed for reproducibility.
    """
    def __init__(self, n_layers=2, kernel='matern', n_inducing_points=100,imputation_strategy='chained',
                 n_samples=20, learning_rate=0.01, n_iterations=10000,
                 batch_size=128, var_noise=0.0001, likelihood_var=0.01,imp_init='mean',
                 use_cuda=True, verbose=True, seed=0):

        self.n_layers = n_layers
        self.kernel = kernel
        self.n_inducing_points = n_inducing_points
        self.n_samples = n_samples
        self.imputation_strategy = imputation_strategy
        self.learning_rate = learning_rate
        self.n_iterations = n_iterations
        self.batch_size = batch_size
        self.var_noise = var_noise
        self.likelihood_var = likelihood_var
        self.use_cuda = use_cuda
        self.seed = seed
        self.imputation_method = imp_init
        # Internal attributes
        self.model_ = None
        self.scaler_ = StandardScaler()
        self.device_ = None
        self.layer_missing_index_ = []
        self.verbose=verbose


    def _initialize_device(self):
        """Initializes the torch device (CPU or CUDA)."""
        if self.use_cuda and torch.cuda.is_available():
            self.device_ = torch.device("cuda")
            torch.cuda.set_device(0) # Assuming first GPU
            torch.backends.cudnn.benchmark = True
        else:
            self.device_ = torch.device("cpu")
        print(f"Using device: {self.device_.type}")

    def _build_holistic_model(self, X_train, y_train, ind_nan):
        """Constructs a single, multi-output DGP model for all features."""
        # Select kernel
        kernel_used = Matern if self.kernel == "matern" else RBF

        Z, lengthscale = self._get_inducing_points_and_lengthscale(X_train)

        # Construct the network layers
        D = X_train.shape[1]
        layer_sizes = []
        kernels = ModuleList()
        likelihood_gaussian = ModuleList()

        for _ in range(self.n_layers):
            kernels.append(
                kernel_used(lengthscales=np.tile(lengthscale, D), variance=100.0, device=self.device_) +
                White(variance=self.var_noise, device=self.device_)
            )
            layer_sizes.append(D)

        num_data = [X_train.shape[0]]
        likelihood_gaussian.append(Gaussian(self.likelihood_var, device=self.device_))

        # Convert data to torch tensors and create the DGP model
        X_train_torch = torch.from_numpy(X_train).to(torch.float64).to(self.device_)
        y_train_torch = torch.from_numpy(y_train).to(torch.float64).to(self.device_)
        Z_torch = torch.from_numpy(Z).to(torch.float64).to(self.device_)

        return DGP(
            X=X_train_torch, Y=y_train_torch, Z=Z_torch, kernels=kernels,
            layer_sizes=layer_sizes, likelihood_gaussian=likelihood_gaussian,
            likelihood_ber=None,
            num_data=np.array(num_data), num_samples=self.n_samples,
            y_mean=torch.from_numpy(self.x_means).to(self.device_, torch.float64),
            y_std=torch.from_numpy(self.x_stds).to(self.device_, torch.float64),
            device=self.device_, ind_total_nan=ind_nan,
            layer_missing_index=self.layer_missing_index_,
            x_means=self.x_means, x_stds=self.x_stds,
            layer=SVGPLayer, use_missing_gp=False
        )

    def _get_inducing_points_and_lengthscale(self, X_train):
        """Helper to calculate inducing points and initial lengthscale."""
        # Create a copy without NaNs for KMeans and distance calculation
        X_train_nonan = np.nan_to_num(X_train, nan=np.nanmean(X_train, axis=0))

        # Determine inducing points using KMeans
        try:
            kmeans = KMeans(n_clusters=self.n_inducing_points, random_state=self.seed, n_init=10).fit(X_train_nonan)
            Z = kmeans.cluster_centers_
        except Exception as e:
            if self.n_inducing_points>= X_train.shape[0]:
                raise exit(f'The number of inducing points {self.n_inducing_points} should be less than data rows {X_train.shape[0]}')
            print(f"KMeans failed: {e}. Falling back to random sampling for inducing points.")
            indices = np.random.choice(X_train.shape[0], self.n_inducing_points, replace=False)
            Z = X_train_nonan[indices, :]

        # Estimate initial lengthscale based on median distance
        if X_train.shape[0] < 1000:
            value = np.expand_dims(np.sum(np.square(X_train_nonan), 1), 1)
            distance = value - 2 * np.matmul(X_train_nonan, np.transpose(X_train_nonan)) + np.transpose(value)
            median = np.percentile(distance, q=50)
        else:
            median = 10.0
        lengthscale = np.log(median) if median > 0 else 1.0

        return Z, lengthscale

    def _build_chained_model(self, X_train, y_train, ind_nan):
        """Constructs a DGP with a separate layer for each feature with missing data."""
        # Select kernel
        kernel_used = Matern if self.kernel == "matern" else RBF

        Z, lengthscale = self._get_inducing_points_and_lengthscale(X_train)

        # Identify and order features with missing values
        total_mis = ind_nan.sum(0)
        features_with_missing = np.where(total_mis > 0)[0]

        # Sort by number of missing values (descending)
        sorted_indices_by_missing = np.argsort(-total_mis[features_with_missing])
        layers_missing_ordered = features_with_missing[sorted_indices_by_missing]

        # Then sort by standard deviation (ascending)
        stds_of_missing_layers = self.x_stds[layers_missing_ordered]
        sorted_indices_by_std = np.argsort(stds_of_missing_layers)
        self.layer_missing_index_ = layers_missing_ordered[sorted_indices_by_std]

        print(f"Chained model processing features in order: {self.layer_missing_index_}")

        num_data = []
        layer_sizes = []
        likelihood_gaussian = ModuleList()
        kernels = torch.nn.ModuleList()

        for layer_idx in self.layer_missing_index_:
            # Each layer predicts one feature from all others
            nD = X_train.shape[1] - 1
            num_data.append((~ind_nan[:, layer_idx]).sum())
            kernels.append(
                kernel_used(lengthscales=np.tile(lengthscale, nD), variance=1.0, device=self.device_) +
                White(variance=self.var_noise, device=self.device_)
            )
            layer_sizes.append(nD)
            likelihood_gaussian.append(Gaussian(self.likelihood_var, device=self.device_))

        if not self.layer_missing_index_.any():
            raise ValueError("Chained strategy selected, but no missing data found.")

        # Convert data to torch tensors and create the DGP model
        X_train_torch = torch.from_numpy(X_train).to(torch.float64).to(self.device_)
        y_train_torch = torch.from_numpy(y_train).to(torch.float64).to(self.device_)
        Z_torch = torch.from_numpy(Z).to(torch.float64).to(self.device_)

        return DGP(
            X=X_train_torch, Y=y_train_torch, Z=Z_torch, kernels=kernels,
            layer_sizes=layer_sizes, likelihood_gaussian=likelihood_gaussian,
            likelihood_ber=None,
            num_data=np.array(num_data), num_samples=self.n_samples,
            y_mean=torch.from_numpy(self.x_means).to(self.device_, torch.float64),
            y_std=torch.from_numpy(self.x_stds).to(self.device_, torch.float64),
            device=self.device_, ind_total_nan=ind_nan,
            layer_missing_index=self.layer_missing_index_,
            x_means=self.x_means, x_stds=self.x_stds,
            layer=SVGPLayer, use_missing_gp=True
        )


    def initial_imputer(self, X):
        if self.imputation_method == 'median':
            imp = SimpleImputer(missing_values=np.nan, strategy='median')
        elif self.imputation_method == 'mean':
            imp = SimpleImputer(missing_values=np.nan, strategy='mean')
        elif self.imputation_method == 'constant':
            imp = SimpleImputer(missing_values=np.nan, strategy='constant', fill_value=0.0)
        elif self.imputation_method == 'mice':
            #lr = BayesianRidge()
            lr = LinearRegression()
            imp = IterativeImputer(estimator=lr, missing_values=np.nan, max_iter=10, verbose=-1,
                                   random_state=42)
        elif self.imputation_method == 'knn':
            imp = KNNImputer(n_neighbors=20, add_indicator=False)
        else:
            raise Exception('Please see the instruction...')
        X = imp.fit_transform(X).astype('float')
        self.init_imputer = imp
        return X

    def compute_mean_std(self, X_train, ind_used_train):
        means, stds = [], []

        for i in range(X_train.shape[1]):
            std = X_train[~ind_used_train[:, i], i].std()
            if std != 0:
                means.append(X_train[~ind_used_train[:, i], i].mean())

                stds.append(std)
            else:
                means.append(0)

                stds.append(0)
        return np.array(means), np.array(stds)

    def fit(self, X, X_preimputed=None):
        """
        Fit the MGP imputer on the data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Input data with missing values, where missing values are encoded as np.nan.

        Returns
        -------
        self : object
            Returns self.
        """
        set_seed(self.seed)
        self._initialize_device()

        # 1. Preprocessing
        X_train = X.copy()
        ind_nan = np.isnan(X_train)
        
        # Fit scaler on the data (ignoring NaNs for mean/std calculation)
        #self.scaler_.fit(X_train)

        #X_scaled = self.scaler_.transform(X_train)
        
        # Replace NaNs with 0 after scaling, as the model expects numerical input.
        # The `ind_nan` mask will tell the model which values were originally missing.
        #X_imputed_mean = np.nan_to_num(X_scaled, nan=0.0)
        if X_preimputed is None:
            X_imputed_mean = self.initial_imputer(X_train)
        else:
            X_imputed_mean = X_preimputed
        x_means, x_stds = self.compute_mean_std(X_imputed_mean, ind_nan)

        self.x_means = (np.median(X_imputed_mean, 0) + np.mean(X_imputed_mean, 0)) / 2.0
        self.x_stds = X_imputed_mean.std(0) + 0.00001

        X_imputed_mean = (X_imputed_mean - self.x_means) / self.x_stds

        # 2. Build Model
        print(f"Building model with '{self.imputation_strategy}' strategy...")
        if self.imputation_strategy == 'holistic':
            self.model_ = self._build_holistic_model(X_imputed_mean, X_imputed_mean, ind_nan)
        elif self.imputation_strategy == 'chained':
            self.model_ = self._build_chained_model(X_imputed_mean, X_imputed_mean, ind_nan)
        else:
            raise ValueError(f"Unknown imputation_strategy: {self.imputation_strategy}")
        
        params = list(self.model_.parameters())
        print(f"Model has {np.sum([p.numel() for p in params])} trainable parameters.")
        
        optimizer = torch.optim.Adam(self.model_.parameters(), lr=self.learning_rate)

        # 3. Training
        # The customLoader needs a "full" dataset for its loss calculation.
        # Since we are doing imputation, the "full" data is the mean-imputed data.

        X_train_full = X_imputed_mean.copy()

        
        # The model expects samples as the first dimension
        X_train_samples = np.tile(np.expand_dims(X_imputed_mean, 0), [self.n_samples, 1, 1])

        training_generator = customLoader(
            X_train_samples,
            X_train_full,
            ind_nan,
            ind_nan,
            self.batch_size,
            shuffle=True,
            device=self.device_,
        )



        n_data_points = X_train.shape[0]
        iter_per_epoch = max(1, n_data_points / self.batch_size)
        epochs = int(np.ceil(self.n_iterations / iter_per_epoch))

        print(f"Starting training for {epochs} epochs...")
        train(
            model=self.model_,
            training_generator=training_generator,
            test_generator=None, # No test set during fitting
            optimizer=optimizer,
            path_file_name=None, # Don't save to file
            batch_size=self.batch_size,
            epochs=epochs,
            device=self.device_,
            layer_missing_index=self.layer_missing_index_,
            lrate=self.learning_rate,
            predict_test=False,
            verbose=self.verbose,
        )
        print("Training complete.")
        return self


    def predict(self, X, batch_size=None, X_preimputed=None, return_prob=False):
        """
        Impute missing values in X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data with missing values to impute.

        Returns
        -------
        X_imputed : array-like, shape (n_samples, n_features)
            Data with missing values imputed.
        """
        if batch_size is None:
            batch_size = self.batch_size
        if self.model_ is None:
            raise RuntimeError("You must call fit before calling predict.")

        X_test = X.copy()
        ind_nan_test = np.isnan(X_test)

        if X_preimputed is None:
            X_imputed_test = self.init_imputer.transform(X_test)
        else:
            X_imputed_test = X_preimputed

        X_imputed_test = (X_imputed_test - self.x_means) / self.x_stds
        # Scale data using the scaler fitted on the training data
        #X_scaled = self.scaler_.transform(X_test)
        #X_imputed_mean = np.nan_to_num(X_scaled, nan=0.0)
        #X_test_full = X_imputed_mean.copy()

        X_test_samples = np.tile(np.expand_dims(X_imputed_test, 0), [self.n_samples, 1, 1])

        # Create a data loader for the test set
        test_generator = customLoader(
            X_test_samples,
            X_imputed_test,
            ind_nan_test,
            ind_nan_test,
            batch_size,
            shuffle=False,
            device=self.device_,
        )
        loggs = logger(
            model=self.model_,
            train_generator=None,
            test_generator=test_generator,
            on_epoch_end_file=None,
            on_train_end_file=None,
            device=self.device_,
            lastlayer=True,
            reset_logger=False,
            verbose=self.verbose,
        )
        # Use the evaluate function to get predictions
        # We only need the test_info part of the output
        _, _, _, _, _, _, _, test_info = evaluate(
            model=self.model_,

            training_generator=None,
            test_generator=test_generator,
            path_file_name=None,
            loggs=loggs,
            batch_size=batch_size,
            device=self.device_,
        )

        _, _, mean_pred, var_pred_aggr_test, ind_nan_aggr_test = test_info

        # Inverse transform to get back to the original data scale
        var_pred_aggr_test = (var_pred_aggr_test * (self.x_stds[None, None, :]) ** 2)
        std_pred_test = np.sqrt(var_pred_aggr_test + 0.0001)
        mean_pred = mean_pred*self.x_stds[None, None, :] + self.x_means[None, None, :]



        # The model returns predictions for all points. We only care about the missing ones.
        # We take the mean of the samples from the posterior.
        mean_imputation_scaled = np.mean(mean_pred, axis=0)
        std_imputatino_scaled = np.mean(std_pred_test, axis=0)
        # Create the final standard deviation
        X_std = np.zeros_like(std_imputatino_scaled)
        X_std[ind_nan_test] = std_imputatino_scaled[ind_nan_test]
        # Create the final imputed dataset
        X_imputed = X.copy()
        X_imputed[ind_nan_test] = mean_imputation_scaled[ind_nan_test]

        if return_prob:
            return X_imputed, X_std, mean_pred, std_pred_test
        else:
            return X_imputed, X_std

    def fit_transform(self, X, X_preimputed=None, **fit_params):
        """
        Fit to data, then transform it.
        This is a convenience method that fits the model and then returns the imputed data.
        """
        return self.fit(X, X_preimputed,**fit_params).predict(X, X_preimputed=X_preimputed)


if __name__ == '__main__':
    # Example Usage
    print("Running MGP Imputer Example...")
    import pandas as pd
    # 1. Create a synthetic dataset with missing values
    np.random.seed(42)

    #n_samples, n_features = 200, 5
    #X_true = np.random.rand(n_samples, n_features) * 10
    #X_missing = X_true.copy()
    # Introduce 20% missing values
    #missing_mask = np.random.rand(n_samples, n_features) < 0.2
    #X_missing[missing_mask] = np.nan

    #print(f"Created a dataset of shape {X_missing.shape} with {np.sum(missing_mask)} missing values.")


    X_true = pd.read_csv('datasets/parkinson.csv').values
    X_missing = pd.read_csv('datasets/parkinson_10.csv').values
    n_samples, n_features = X_true.shape
    missing_mask = np.isnan(X_missing.astype(np.float64)) > 0


    # 2. Initialize and fit the imputer
    # Using fewer iterations for a quick example run
    mgp_imputer = MGPImputer(
        n_layers=1,
        n_inducing_points=100,
        n_iterations=500,
        learning_rate=0.01,
        batch_size=100,
        imputation_strategy='holistic',#holistic or chained
        imp_init='mean',
        use_cuda=True,
        seed=42
    )

    # 3. Fit on the data with missing values and then transform it
    X_imputed, X_std_imputed = mgp_imputer.fit_transform(X_missing, X_preimputed=None)

    # 4. Evaluate the imputation
    rmse = np.sqrt(np.mean((X_imputed[missing_mask] - X_true[missing_mask])**2))
    print(f"\nImputation complete.")
    print(f"RMSE on missing values: {rmse:.4f}")

    # You can now use the `mgp_imputer` to predict on new data
    # For example, create a new test set with missing values
    #X_test_true = np.random.rand(50, n_features) * 10
    #X_test_missing = X_test_true.copy()
    #test_missing_mask = np.random.rand(50, n_features) < 0.2
    #X_test_missing[test_missing_mask] = np.nan

    X_test_true = pd.read_csv('datasets/parkinson_test.csv').values
    X_test_missing = pd.read_csv('datasets/parkinson_test_10.csv').values
    test_missing_mask = np.isnan(X_test_missing.astype(np.float64)) > 0


    X_test_imputed, X_std_imputed = mgp_imputer.predict(X_test_missing)
    test_rmse = np.sqrt(np.mean((X_test_imputed[test_missing_mask] - X_test_true[test_missing_mask])**2))
    print(f"RMSE on a new test set: {test_rmse:.4f}")




from tabnanny import verbose

import torch
import numpy as np
import time

from mgp.model.utils import *
from mgp.model.callback import logger
import pkbar
from torch.optim.lr_scheduler import StepLR


def train(
    model,
    training_generator,
    test_generator,
    optimizer,
    path_file_name,
    batch_size=100,
    epochs=100,
    device=None,
    layer_missing_index=None,
    lrate=0.1,
    ind_binary=None,
    predict_test=True,
    verbose=False,
):
    """
    @param model: model
    @param training_generator: training data
    @param test_generator: test data
    @param optimizer: optimizer
    @param path_file_name: path for writing the output
    @param batch_size: batch size
    @param epochs: number of epochs
    @param device: the used device CPU/GPU
    @param layer_missing_index: missign index for the layers
    @param lrate: learning rate
    @param ind_binary:
    @param predict_test: prediction on test
    @return: trained model
    """

    model.train()

    if batch_size is None:
        batch_size = model.training_generator.model.batch_size
    length_dataset = len(training_generator.dataset)
    assert batch_size <= length_dataset
    if path_file_name is not None:
        path_results = path_file_name[: path_file_name.rfind("/")]
        one_train_end_file = (
            path_results
            + "/"
            + path_file_name[path_file_name.rfind("/") + 1 :][0:-4]
            + "_final.txt"
        )
    else:
        one_train_end_file = None
        path_results = None
    missing_pred = False

    if epochs > 1:
        reset_logger = True
    else:
        reset_logger = False
    if one_train_end_file is None:
        reset_logger = False
    loggs = logger(
        model=model,
        train_generator=training_generator,
        test_generator=test_generator,
        on_epoch_end_file=path_file_name,
        on_train_end_file=one_train_end_file,
        device=device,
        lastlayer=not missing_pred,
        reset_logger=reset_logger,
        verbose=verbose,
    )
    if epochs == 0:
        return loggs

    # Generate variables and operations for the minimizer and initialize variables
    global_nelbow = np.inf
    train_per_epoch = len(training_generator.dataloader)

    ini_time = time.time()

    totla_time = 0.0

    num_divide = 1

    BEST_NELBO = np.inf
    optimizer = torch.optim.Adam(model.parameters(), lr=lrate, weight_decay=0.00)

    for el in range(num_divide):
        for param in model.parameters():
            param.requires_grad = True

        for e in range(epochs):
            message_write = {}
            message_write["epoch"] = e
            if verbose:
                kbar = pkbar.Kbar(
                    target=train_per_epoch,
                    epoch=e,
                    num_epochs=epochs,
                    width=50,
                    always_stateful=False,
                )

            avg_acc_c = 0
            avg_nll_c = 0

            start = time.time()

            idx = 0
            #if verbose:
            #    kbar.update(
            #        idx,
            #        values=[
            #            ("nelbo", 0),
                        #("rmse", 0),
                        #("nll", 0),
             #           ("rmse_c", 0),
             #           ("nll_c", 0),
             #       ],
             #   )
            avg_nelbo = 0.0
            avg_acc = 0
            avg_nll = 0

            idx += 1

            for i, data in enumerate(training_generator):
                batch_x, batch_y, batch_ind_nan, batch_ind_nan_target = data
                if batch_x.ndim == 3:
                    batch_x = np.transpose(batch_x, (1, 0, 2))

                loss = model.train_step(
                    optimizer,
                    batch_x.to(device),
                    batch_y.to(device),
                    batch_ind_nan,
                    batch_ind_nan_target,
                    e * train_per_epoch + i,
                    np.inf,
                )

                if i == 0:
                    y_aggr = model.y_aggr
                    mean_pred_aggr = model.mean_pred_aggr
                    std_pred_aggr = model.std_pred_aggr

                else:
                    if model.use_missing_gp:
                        y_aggr = torch.vstack((y_aggr, model.y_aggr))
                        mean_pred_aggr = torch.cat(
                            (mean_pred_aggr, model.mean_pred_aggr), 1
                        )
                        std_pred_aggr = torch.cat(
                            (std_pred_aggr, model.std_pred_aggr), 1
                        )
                    else:
                        y_aggr = torch.cat((y_aggr, model.y_aggr), 0)
                        mean_pred_aggr = torch.cat(
                            (mean_pred_aggr, model.mean_pred_aggr), 1
                        )
                        std_pred_aggr = torch.cat(
                            (std_pred_aggr, model.std_pred_aggr), 1
                        )
                acc = model.rmse_val
                nll = model.nll_val
                acc_c = model.rmse_val_converted
                nll_c = model.nll_val_converted
                avg_acc_c += acc_c
                avg_nll_c += nll_c
                avg_acc += acc
                avg_nll += nll
                avg_nelbo += loss
                if verbose:
                    kbar.update(
                        idx,
                        values=[
                            ("nelbo", loss),
                            #("rmse", acc),
                            #("nll", nll),
                            ("rmse_c", acc_c),
                            ("nll_c", nll_c),
                        ],
                    )
                idx += 1

            NELB, NLL, ACC, NLL_c, ACC_c = (
                avg_nelbo / train_per_epoch,
                avg_nll / train_per_epoch,
                avg_acc / train_per_epoch,
                avg_nll_c / train_per_epoch,
                avg_acc_c / train_per_epoch,
            )

            (
                message_write["acc_train"],
                message_write["nelbo_train"],
                message_write["nll_train"],
            ) = (
                ACC.detach().cpu().numpy().item(),
                NELB.detach().cpu().numpy().item(),
                NLL.detach().cpu().numpy().item(),
            )

            message_write["acc_train_c"], message_write["nll_train_c"] = (
                ACC_c.detach().cpu().numpy().item(),
                NLL_c.detach().cpu().numpy().item(),
            )

            if NELB < BEST_NELBO:
                BEST_NELBO = NELB
                if path_results is not None:
                    if model.use_missing_gp:
                        save_model(
                            model, optimizer, device, path_results, epoch="latest_missing"
                        )
                    else:
                        save_model(model, optimizer, device, path_results)

                    start_write_results = time.time()
                    ini_time = loggs.on_epoch_end(
                        ini_time, message_write, predict_test=predict_test
                    )

                    end_write_results = time.time()
                    write_time = end_write_results - start_write_results

                    if NELB < global_nelbow:
                        start_save_time = time.time()

                        end_save_time = time.time()
                        save_time = end_save_time - start_save_time
                        ini_time += save_time
                        global_nelbow = NELB

                        end = time.time() - save_time - write_time
                    else:
                        end = time.time() - write_time

                    totla_time += end - start
    """
    
    if model.use_missing_gp:
        load_model(
            model, optimizer, path_results, which_epoch="latest_missing", device=device
        )
    else:
        load_model(model, optimizer, path_results, which_epoch="latest", device=device)
    """
    # NELBO_train, NLL_train, ACC_train, NELBO_test, NLL_test, ACC_test, y_mst_train, y_mst_test = loggs.on_train_end()
    return loggs


def evaluate(
    model,
    training_generator,
    test_generator,
    path_file_name,
    loggs,
    batch_size=100,
    device=None,
):
    model.eval()

    #if batch_size is None:
    #    batch_size = model.training_generator.model.batch_size
    #length_dataset = len(test_generator.dataset)
    #assert batch_size <= length_dataset
    if path_file_name is not None:
        path_results = path_file_name[: path_file_name.rfind("/")]
        one_train_end_file = (
            path_results
            + "/"
            + path_file_name[path_file_name.rfind("/") + 1 :][0:-4]
            + "_final.txt"
        )
    else:
        path_results = None

    (
        NELBO_train,
        NLL_train,
        ACC_train,
        NELBO_test,
        NLL_test,
        ACC_test,
        train_info,
        test_info,
    ) = loggs.on_eval(training_generator, test_generator)

    return (
        NELBO_train,
        NLL_train,
        ACC_train,
        NELBO_test,
        NLL_test,
        ACC_test,
        train_info,
        test_info,
    )



import torch

# import numpy as np
# torch.multiprocessing.set_start_method('spawn')


class CDataSet(torch.utils.data.Dataset):
    def __init__(self, X, y, device="cpu", dtype=torch.float64):
        self.X = X
        self.y = y
        self.device = device
        self.dtype = dtype

    def __len__(self):
        return len(self.y)

    def __getitem__(self, item):
        X = self.X[item, :]
        y = self.y[item, :]
        x = torch.from_numpy(X).to(self.dtype)
        y = torch.from_numpy(y).to(self.dtype)
        return x, y[0]



import csv
from tabnanny import verbose

import torch
import time
import pkbar
from mgp.model.utils import *
import pandas as pd
from os.path import join, dirname
import numpy as np


class logger:
    def __init__(
        self,
        model,
        train_generator,
        test_generator,
        on_epoch_end_file,
        on_train_end_file,
        device=None,
        lastlayer=False,
        reset_logger=True,
        verbose=False,
    ):
        self.on_epoch_end_file = on_epoch_end_file
        self.on_train_end_file = on_train_end_file
        self.train_generator = train_generator
        self.test_generator = test_generator
        self.total_train_time = 0.0
        self.model = model
        self.verbose = verbose
        self.lastlayer = lastlayer
        self.device = device
        if reset_logger:
            with open(on_epoch_end_file, "w") as myfile:
                log_writer = csv.writer(myfile)
                log_writer.writerow(
                    [
                        "epoch",
                        "nelbo_train",
                        "acc_train",
                        "nll_train",
                        "acc_train_c",
                        "nll_train_c",
                        "nelbo_test",
                        "acc_test",
                        "nll_test",
                        "acc_test_c",
                        "nll_test_c",
                        "total_time_train",
                        "prediction_time",
                    ]
                )
            with open(on_train_end_file, "w") as myfile:
                log_writer = csv.writer(myfile)
                log_writer.writerow(
                    [
                        "RMSE_train",
                        "NLL_train",
                        "RMSE_train_c",
                        "NLL_train_c",
                        "RMSE_test",
                        "NLL_test",
                        "RMSE_test_c",
                        "NLL_test_c",
                        "total_training_time",
                        "prediction_time",
                    ]
                )

    def evaluate(self, generator, normalize_y=True):
        with torch.no_grad():
            _per_epoch = len(generator.dataloader)
            if self.verbose:
                kbar = pkbar.Kbar(target=_per_epoch, width=50, always_stateful=False)
                kbar.update(
                    0,
                    values=[
                        ("nelbo", 0),
                        #("rmse", 0),
                        #("nll", 0),
                        ("rmse_c", 0),
                        ("nll_c", 0),
                    ],
                )
            # idx += 1
            avg_nelbo = 0
            avg_acc = 0
            avg_nll = 0
            avg_acc_c = 0
            avg_nll_c = 0
            for idx, data in enumerate(generator):
                batch_x, batch_y, batch_ind_nan, batch_ind_nan_target = data
                if batch_x.ndim == 3:
                    batch_x = np.transpose(batch_x, (1, 0, 2))

                loss = self.model.test_step(
                    batch_x.to(self.device),
                    batch_y.to(self.device),
                    batch_ind_nan,
                    batch_ind_nan_target,
                    normalize_y=True,
                )

                if idx == 0:
                    y_aggr = self.model.y_aggr
                    x_aggr = batch_x
                    mean_pred_aggr = self.model.mean_pred_aggr
                    std_pred_aggr = self.model.std_pred_aggr
                    ind_nan_aggr = batch_ind_nan

                else:
                    if self.model.use_missing_gp:
                        y_aggr = torch.vstack((y_aggr, self.model.y_aggr))
                        x_aggr = torch.cat((x_aggr, batch_x), 1)
                        mean_pred_aggr = torch.cat(
                            (mean_pred_aggr, self.model.mean_pred_aggr), 1
                        )
                        std_pred_aggr = torch.cat(
                            (std_pred_aggr, self.model.std_pred_aggr), 1
                        )
                        ind_nan_aggr = torch.cat((ind_nan_aggr, batch_ind_nan), 0)

                    else:
                        y_aggr = torch.vstack((y_aggr, self.model.y_aggr))
                        x_aggr = torch.cat((x_aggr, batch_x), 1)
                        mean_pred_aggr = torch.cat(
                            (mean_pred_aggr, self.model.mean_pred_aggr), 1
                        )
                        std_pred_aggr = torch.cat(
                            (std_pred_aggr, self.model.std_pred_aggr), 1
                        )
                        ind_nan_aggr = torch.cat((ind_nan_aggr, batch_ind_nan), 0)

                avg_nelbo += loss.detach().cpu().numpy()
                acc = self.model.rmse_val
                nll = self.model.nll_val
                acc_c = self.model.rmse_val_converted
                nll_c = self.model.nll_val_converted
                avg_acc += acc.detach().cpu().numpy()
                avg_nll += nll.detach().cpu().numpy()
                avg_acc_c += acc_c.detach().cpu().numpy()
                avg_nll_c += nll_c.detach().cpu().numpy()
                if self.verbose:
                    kbar.update(
                        idx + 1,
                        values=[
                            ("nelbo", loss),
                            #("rmse", acc),
                            #("nll", nll),
                            ("rmse_c", acc_c),
                            ("nll_c", nll_c),
                        ],
                    )
                # idx += 1
        NLL, ACC = compute_nn_mse(y_aggr, mean_pred_aggr, std_pred_aggr, ind_nan_aggr)
        NLL = NLL.detach().cpu().numpy()
        ACC = ACC.detach().cpu().numpy()
        NLL_c, ACC_c = compute_nn_mse(
            y_aggr * self.model.y_std + self.model.y_mean,
            mean_pred_aggr * self.model.y_std + self.model.y_mean,
            std_pred_aggr * self.model.y_std,
            ind_nan_aggr,
        )
        NLL_c = NLL_c.detach().cpu().numpy()
        ACC_c = ACC_c.detach().cpu().numpy()
        #print(ACC, NLL, ACC_c, NLL_c)
        #print("\n")

        y_m_std = torch.stack((mean_pred_aggr, std_pred_aggr)).transpose(0, 1)

        NELBO, _, _ = avg_nelbo / _per_epoch, avg_nll / _per_epoch, avg_acc / _per_epoch

        return (
            NELBO,
            NLL,
            ACC,
            NLL_c,
            ACC_c,
            [x_aggr, y_aggr, mean_pred_aggr, std_pred_aggr, ind_nan_aggr],
        )

    def on_epoch_end(self, ini_time, message_write={}, predict_test=True):
        """
        write the results to the file the firs columns is RMSE, scond one is Standaridzed RMSE, third one NLL, last one time
        @param means: computed posterior means
        @param variances: computed posterior variances
        @param output: target points
        @param ini_time: time start
        @param file_name: output file path and name
        @return: nothing
        """
        start_predict = time.time()
        NELBO, NLL, ACC, NLL_c, ACC_c = 0.0, 0.0, 0.0, 0.0, 0.0
        self.model.eval()

        if predict_test:
            NELBO, NLL, ACC, NLL_c, ACC_c, _ = self.evaluate(
                self.test_generator, normalize_y=False
            )
        end_predict = time.time()
        prediction_time = end_predict - start_predict
        ini_time += prediction_time
        print("\n")
        self.total_train_time = time.time() - ini_time
        with open(self.on_epoch_end_file, "a") as myfile:
            log_writer = csv.writer(myfile)
            log_writer.writerow(
                [
                    message_write["epoch"],
                    message_write["nelbo_train"],
                    message_write["acc_train"],
                    message_write["nll_train"],
                    message_write["acc_train_c"],
                    message_write["nll_train_c"],
                    NELBO,
                    ACC,
                    NLL,
                    ACC_c,
                    NLL_c,
                    self.total_train_time,
                    prediction_time,
                ]
            )

        self.model.train()

        return ini_time

    def on_train_end(self, training_gen=None, test_gen=None):
        pass

    def on_eval(self, training_gen=None, test_gen=None):
        start_predict = time.time()
        self.model.eval()

        if training_gen:
            self.train_generator = training_gen
        if test_gen:
            self.test_generator = test_gen

        if not self.model.use_missing_gp:
            end_predict = time.time()
            (
                NELBO_test,
                NLL_test,
                ACC_test,
                NLL_test_c,
                ACC_test_c,
                [
                    x_aggr_test,
                    y_aggr_test,
                    mean_pred_aggr_test,
                    std_pred_aggr_test,
                    ind_nan_aggr_test,
                ],
            ) = self.evaluate(self.test_generator, normalize_y=False)

            x_aggr_test = x_aggr_test.detach().cpu().numpy()
            y_aggr_test = y_aggr_test.detach().cpu().numpy()
            mean_pred_aggr_test = mean_pred_aggr_test.detach().cpu().numpy()
            std_pred_aggr_test = std_pred_aggr_test.detach().cpu().numpy()
            ind_nan_aggr_test = ind_nan_aggr_test.detach().cpu().numpy()
            print("\n")
            if self.train_generator is not None:
                (
                    NELBO_train,
                    NLL_train,
                    ACC_train,
                    NLL_train_c,
                    ACC_train_c,
                    [
                        x_aggr_train,
                        y_aggr_train,
                        mean_pred_aggr_train,
                        std_pred_aggr_train,
                        ind_nan_aggr_train,
                    ],
                ) = self.evaluate(self.train_generator, normalize_y=False)

                if torch.is_tensor(mean_pred_aggr_train):
                    y_aggr_train = y_aggr_train.detach().cpu().numpy()
                    x_aggr_train = x_aggr_train.detach().cpu().numpy()
                    mean_pred_aggr_train = mean_pred_aggr_train.detach().cpu().numpy()
                    std_pred_aggr_train = std_pred_aggr_train.detach().cpu().numpy()
                    ind_nan_aggr_train = ind_nan_aggr_train.detach().cpu().numpy()


            else:
                NELBO_train = None,
                NLL_train = None,
                ACC_train = None,
                NLL_train_c = None,
                ACC_train_c = None,
                x_aggr_train = None,
                y_aggr_train = None,
                mean_pred_aggr_train = None,
                std_pred_aggr_train = None,
                ind_nan_aggr_train = None,


        else:
            end_predict = time.time()
            if self.train_generator is not None:
                (
                    NELBO_train,
                    NLL_train,
                    ACC_train,
                    NLL_train_c,
                    ACC_train_c,
                    [
                        x_aggr_train,
                        y_aggr_train,
                        mean_pred_aggr_train,
                        std_pred_aggr_train,
                        ind_nan_aggr_train,
                    ],
                ) = self.evaluate(self.train_generator, normalize_y=False)
                if torch.is_tensor(mean_pred_aggr_train):
                    y_aggr_train = y_aggr_train.detach().cpu().numpy()
                    x_aggr_train = x_aggr_train.detach().cpu().numpy()
                    mean_pred_aggr_train = mean_pred_aggr_train.detach().cpu().numpy()
                    std_pred_aggr_train = std_pred_aggr_train.detach().cpu().numpy()
                    ind_nan_aggr_train = ind_nan_aggr_train.detach().cpu().numpy()

            else:
                NELBO_train = None,
                NLL_train = None,
                ACC_train=None,
                NLL_train_c=None,
                ACC_train_c=None,
                x_aggr_train = None,
                y_aggr_train = None,
                mean_pred_aggr_train = None,
                std_pred_aggr_train = None,
                ind_nan_aggr_train = None,
            (
                NELBO_test,
                NLL_test,
                ACC_test,
                NLL_test_c,
                ACC_test_c,
                [
                    x_aggr_test,
                    y_aggr_test,
                    mean_pred_aggr_test,
                    std_pred_aggr_test,
                    ind_nan_aggr_test,
                ],
            ) = self.evaluate(self.test_generator, normalize_y=False)
            if torch.is_tensor(mean_pred_aggr_test):
                x_aggr_test = x_aggr_test.detach().cpu().numpy()
                y_aggr_test = y_aggr_test.detach().cpu().numpy()
                mean_pred_aggr_test = mean_pred_aggr_test.detach().cpu().numpy()
                std_pred_aggr_test = std_pred_aggr_test.detach().cpu().numpy()
                ind_nan_aggr_test = ind_nan_aggr_test.detach().cpu().numpy()


        NELBO_test = 0.0
        prediction_time = end_predict - start_predict
        """
        
        if self.verbose:
            print("Average likelihood {:.4f}".format(ACC_test))
            print("Average rmse {:.4f}".format(NLL_test))
            print(
                "prediction time {}".format( prediction_time
                )
            )
        """
        #outfile = join(dirname(self.on_train_end_file), "results_missing_pred_test.csv")
        if self.on_train_end_file is not None:
            with open(self.on_train_end_file, "a") as myfile:
                log_writer = csv.writer(myfile)
                log_writer.writerow(
                    [
                        ACC_train,
                        NLL_train,
                        ACC_train_c,
                        NLL_train_c,
                        ACC_test,
                        NLL_test,
                        ACC_test_c,
                        NLL_test_c,
                        self.total_train_time,
                        prediction_time,
                    ]
                )
        return (
            NELBO_train,
            NLL_train,
            ACC_train,
            NELBO_test,
            NLL_test,
            ACC_test,
            [
                x_aggr_train,
                y_aggr_train,
                mean_pred_aggr_train,
                std_pred_aggr_train,
                ind_nan_aggr_train,
            ],
            [
                x_aggr_test,
                y_aggr_test,
                mean_pred_aggr_test,
                std_pred_aggr_test,
                ind_nan_aggr_test,
            ],
        )



import numpy as np
import pandas as pd
from sklearn.impute import SimpleImputer
#from fancyimpute import KNN
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer
from sklearn.linear_model import LinearRegression, BayesianRidge
from sklearn.preprocessing import (
    MinMaxScaler,
    MaxAbsScaler,
    StandardScaler,
    RobustScaler,
)
from sklearn.impute import KNNImputer
from sklearn.metrics import mean_squared_error
#from rpy2.robjects.packages import importr
#import rpy2.robjects as ro
#import rpy2.robjects.numpy2ri

#rpy2.robjects.numpy2ri.activate()


def mean_imputer(x, ind_total_nan):
    for o in range(x.shape[1]):
        ind_true = ~ind_total_nan[:, o]
        uq = np.mean(x[ind_true, o])
        x[ind_total_nan[:, o], :][:, o] = uq

    return x


def load_dataset(opt):
    ######################## inputs ###############################

    dta = opt.dataset_name.split("_")[0]  #'clinicaldata'

    df_train_missing = pd.read_csv(
        "datasets/{}/s{}/{}_train_data.csv".format(
            opt.dataset_name, opt.split_number, dta
        )
    )
    df_train_lables = pd.read_csv(
        "datasets/{}/s{}/{}_train_labels.csv".format(
            opt.dataset_name, opt.split_number, dta
        ),
    )

    df_train_ = pd.concat((df_train_missing, df_train_lables), axis=1)
    df_add_missing = pd.read_csv(
        "datasets/{}/s{}/{}_test_data.csv".format(
            opt.dataset_name, opt.split_number, dta
        ),
    )
    df_test_labels = pd.read_csv(
        "datasets/{}/s{}/{}_test_labels.csv".format(
            opt.dataset_name, opt.split_number, dta
        ),
    )
    df_test_ = pd.concat((df_add_missing, df_test_labels), axis=1)

    df_train_full = pd.read_csv(
        "datasets/{}/s{}/{}_train_data.csv".format(dta, opt.split_number, dta)
    )
    df_train_full_l = pd.read_csv(
        "datasets/{}/s{}/{}_train_labels.csv".format(dta, opt.split_number, dta)
    )
    df_train_full = pd.concat((df_train_full, df_train_full_l), axis=1)

    df_test_full = pd.read_csv(
        "datasets/{}/s{}/{}_test_data.csv".format(dta, opt.split_number, dta),
    )
    df_test_full_l = pd.read_csv(
        "datasets/{}/s{}/{}_test_labels.csv".format(dta, opt.split_number, dta),
    )
    df_test_full = pd.concat((df_test_full, df_test_full_l), axis=1)

    if "id" in df_train_.columns:
        df_train = df_train_.drop("id", axis=1)
        df_test = df_test_.drop("id", axis=1)
    else:
        df_train = df_train_
        df_test = df_test_

    df_current_train = df_train
    df_current_test = df_test

    global ind_binary
    ind_binary = []
    r = 0
    for col in df_current_train:
        uq = [u for u in df_current_train[col].unique() if u == u]
        if len(uq) == 2:
            ind_binary.append(r)
        r += 1
    ind_binary = []
    cols = df_current_train.columns

    if opt.nolayers == np.inf:
        X_test = df_current_test.values.astype("float")
        X_train = df_current_train.values.astype("float")
        y_train = None
        y_test = None
    else:
        df_current_train = df_current_train.replace("?", np.nan)
        df_current_test = df_current_test.replace("?", np.nan)

        X_train = df_current_train.values.astype("float")
        X_test = df_current_test.values.astype("float")
        y_train = df_current_train[cols[-1]].values.astype("float")
        y_test = df_current_test[cols[-1]].values.astype("float")

        df_train_full = df_train_full.replace("?", np.nan)
        df_test_full = df_test_full.replace("?", np.nan)
        X_train_full = df_train_full.values.astype("float")
        X_test_full = df_test_full.values.astype("float")

    if opt.imputation == "median":
        imp = SimpleImputer(missing_values=np.nan, strategy="median")
    elif opt.imputation == "mean":
        imp = SimpleImputer(missing_values=np.nan, strategy="mean")
    elif opt.imputation == "constant":
        imp = SimpleImputer(missing_values=np.nan, strategy="constant", fill_value=0.0)
    elif opt.imputation == "mice":
        lr = LinearRegression()
        imp = IterativeImputer(
            estimator=lr, missing_values=np.nan, max_iter=10, verbose=2, random_state=0
        )
    elif opt.imputation == "linear":
        lr = LinearRegression()
        imp = IterativeImputer(
            estimator=lr, missing_values=np.nan, max_iter=1, verbose=2, random_state=0
        )

    elif opt.imputation == "knn":
        imp = KNNImputer(n_neighbors=2, add_indicator=False)
    else:
        raise Exception("Please see the instruction...")

    if opt.nolayers == np.inf:
        ind_n_test = np.isnan(X_test)
        for ir in range(ind_n_test.shape[1]):
            indnan = ind_n_test[:, ir]
            ind_true = ~indnan
            uq = np.unique(X_test[ind_true, ir])
            if indnan.sum() > 0:
                if len(uq) <= 4:
                    X_test[indnan, ir] = np.mean(uq)
                    X_train[indnan, ir] = np.mean(uq)
                else:
                    X_test[indnan, ir] = np.mean(X_test[ind_true, ir])
                    X_train[indnan, ir] = np.mean(X_test[ind_true, ir])

    if opt.nolayers == np.inf:
        ind_total_nan = np.isnan(df_current_train.values.astype(np.float64)) > 0
        ind_total_nan = ind_total_nan * ~ind_n_test
        rmse_inint = np.sqrt(
            mean_squared_error(X_train[ind_total_nan], X_test[ind_total_nan])
        )
        print("rMSE INIT {}, {}".format(opt.imputation, rmse_inint))
    else:
        ind_total_nan = np.isnan(df_current_train.values.astype(np.float64)) > 0

    global ind_total_nan_test

    ind_total_nan_test = np.isnan(df_current_test.values.astype(np.float64)) > 0
    ind_used_train = df_current_train.isna().values
    if opt.imputation != "rmean":
        X_train = imp.fit_transform(X_train).astype("float")
        X_test = imp.transform(X_test).astype("float")
        X_test_full = imp.transform(X_test_full).astype("float")
        X_train_full = imp.transform(X_train_full).astype("float")

    else:
        X_train = mean_imputer(X_train, ind_total_nan)

    x_means, x_stds = scaler_fit(X_train, ind_used_train)
    if opt.nolayers == np.inf:
        X_train = scaler_transform(x_means, x_stds, X_train)
        X_test = scaler_transform(x_means, x_stds, X_test)
        scaler = None
    else:
        x_means = (np.median(X_train, 0) + np.mean(X_train, 0)) / 2.0
        x_stds = X_train.std(0) + 0.00001
        rmse = []
        for d in range(ind_total_nan_test.shape[1]):
            if ind_total_nan_test[:, d].sum() > 0:
                a = np.sqrt(
                    np.mean(
                        (
                            X_test[ind_total_nan_test[:, d], d]
                            - X_test_full[ind_total_nan_test[:, d], d]
                        )
                        ** 2
                    )
                )
                rmse.append(a)
        rmse = np.stack(rmse)
        print("mean imputation: ", np.mean(rmse))

        X_train = (X_train - x_means) / x_stds
        X_test = (X_test - x_means) / x_stds
        X_train_full = (X_train_full - x_means) / x_stds
        X_test_full = (X_test_full - x_means) / x_stds
        rmse = []
        for d in range(ind_total_nan_test.shape[1]):
            if ind_total_nan_test[:, d].sum() > 0:
                a = np.sqrt(
                    np.mean(
                        (
                            X_test[ind_total_nan_test[:, d], d]
                            - X_test_full[ind_total_nan_test[:, d], d]
                        )
                        ** 2
                    )
                )
                rmse.append(a)
        rmse = np.stack(rmse)
        print("mean imputation: ", np.mean(rmse))

        scaler = None  # RobustScaler()

    if opt.nolayers != np.inf:
        y_mean = np.mean(y_train).reshape(1)
        y_std = np.std(y_train).reshape(1)
        y_train = (y_train - y_mean) / y_std
        y_test = (y_test - y_mean) / y_std
    else:
        y_mean, y_std, y_train, y_test = None, None, None, None
    return [
        X_train,
        X_test,
        ind_total_nan,
        imp,
        ind_binary,
        x_means,
        x_stds,
        ind_total_nan_test,
        scaler,
        y_train,
        y_test,
        y_mean,
        y_std,
        X_train_full,
        X_test_full,
    ]


def scaler_fit(X_train, ind_used_train):
    means, stds = [], []

    for i in range(X_train.shape[1]):
        std = X_train[~ind_used_train[:, i], i].std()
        if std != 0:
            means.append(X_train[~ind_used_train[:, i], i].mean())

            stds.append(std)
        else:
            means.append(0)

            stds.append(0)
    return means, stds


def scaler_transform(means, stds, X_train):
    for i in range(X_train.shape[1]):
        if stds[i] != 0:
            X_train[:, i] = (X_train[:, i] - means[i]) / stds[i]

    return X_train


def prepare_data(df_current_train, df_current_test, target_ind="VCT_"):
    print("preprocessing data ...")
    for el in df_current_train:
        if el == "id":
            continue
        if df_current_train[el].dtype == "O":
            list_uqs = list(
                set(
                    np.hstack(
                        [df_current_test[el].unique(), df_current_train[el].unique()]
                    )
                )
            )
            uq = [u for u in list_uqs if u == u]
            imputed_values = [0, 1]
            uq = np.sort(uq)
            if len(uq) == 2:
                # df_current_train[el] = df_current_train[el].astype('int')
                len_train = df_current_train.shape[0]
                dummies = pd.get_dummies(
                    pd.concat([df_current_train[el], df_current_test[el]], 0),
                    drop_first=True,
                ).astype("object")
                dummies.columns = [
                    el + "_" + col for col in list(dummies.columns.astype("str"))
                ]
                ind_n = (
                    pd.concat([df_current_train[el], df_current_test[el]], 0)
                    .isna()
                    .values
                )
                for col in dummies.columns:
                    dummies[col][dummies[col] == 0] = 0
                    dummies[col][dummies[col] == 1] = 1
                    dummies[col][ind_n] = np.nan
                df_current_train = pd.concat([df_current_train, dummies[:len_train]], 1)
                df_current_test = pd.concat([df_current_test, dummies[len_train:]], 1)
                df_current_train = df_current_train.drop(el, 1)
                df_current_test = df_current_test.drop(el, 1)
                ls = list(df_current_train.columns)
                ls.remove("VCT_")
                ls.append("VCT_")
                df_current_train = df_current_train[ls]
                df_current_test = df_current_test[ls]

    return df_current_train, df_current_test  # , means, stds



import torch.nn as nn
import numpy as np
import torch
from mgp.model.utils import reparameterize


default_jitter = 1e-7


class Layer(nn.Module):
    def __init__(self, input_prop_dim=None, dtype=None, device=None, **kwargs):
        """ """
        super().__init__()
        self.device = device
        self.dtype = dtype
        self.input_prop_dim = input_prop_dim

    def conditional_ND(self, X, full_cov=False):
        raise NotImplementedError

    def KL(self):
        pass

    def conditional_SND(self, X, full_cov=False, ind_nan=None):
        """
        @param X: input matrix
        @param full_cov: boolean full covariance
        @param ind_nan: indices of missing values
        @return: multisample conditional mean and variance
        """
        if not full_cov:
            S, N, D = X.shape

            X_flat = X.reshape([S * N, D])
            mean, var = self.conditional_ND(X_flat, ind_nan=ind_nan)
            # num_outputs = tf.shape(mean)[-1]
            num_outputs = self.num_outputs
            return [m.reshape([S, N, num_outputs]) for m in [mean, var]]

    def sample_from_conditional(
        self,
        X,
        z=None,
        full_cov=False,
        max_use=None,
        ind_binary=None,
        training=True,
        ind_nan=None,
    ):
        """
        @param X: input matrix
        @param z: matrix for sampling
        @param full_cov: boolean
        @param max_use: boolean
        @param ind_binary:
        @param training: boolean
        @param ind_nan: indice missing values
        @return:
        """
        mean, var = self.conditional_SND(X, full_cov=full_cov, ind_nan=ind_nan)

        # set shapes
        S, N, _ = X.shape
        D = mean.shape[-1]

        mean = mean.reshape((S, N, D))
        if full_cov:
            var = var.reshape((S, N, N, D))
        else:
            var = var.reshape((S, N, D))

        if z is None:
            z = torch.randn((mean.shape), dtype=self.dtype, device=self.device)

        samples = reparameterize(mean, var, z, full_cov=full_cov)

        return samples, mean, var


class SVGPLayer(Layer):
    def __init__(
        self,
        kern,
        Z,
        num_outputs,
        mean_function,
        white=False,
        input_prop_dim=None,
        seed=0,
        dtype=torch.float64,
        device=None,
        ind_total_nan=None,
        n_layers=None,
    ):
        """ """
        super().__init__(input_prop_dim=input_prop_dim, dtype=dtype, device=device)
        self.seed = seed

        self.num_inducing = Z.shape[0]

        # Inducing points prior mean
        self.register_parameter(
            "q_mu",
            nn.Parameter(
                torch.tensor(
                    np.zeros((self.num_inducing, num_outputs)),
                    dtype=self.dtype,
                    device=self.device,
                )
            ),
        )

        self.register_parameter(
            "inducing_points", nn.Parameter(Z.clone().detach().requires_grad_(True))
        )

        self.kern = kern
        self.mean_function = mean_function

        self.num_outputs = num_outputs

        self.white = white
        self.ind_total_nan = ind_total_nan

        if not self.white:  # initialize to prior
            Ku = self.kern.K(Z)
            Lu = torch.linalg.cholesky(
                Ku + torch.eye(Z.shape[0]).to(self.device) * default_jitter
            )
            Lu_tiled = torch.tile(Lu[None, :, :], [num_outputs, 1, 1])
            li, lj = torch.tril_indices(self.num_inducing, self.num_inducing, 0)
            triangular_Lu = Lu_tiled[:, li, lj]

            self.register_parameter(
                "q_sqrt_tri",
                nn.Parameter(triangular_Lu.clone().detach().requires_grad_(True)),
            )

    def get_prior_matrices(self):
        """ " """
        Ku = self.kern.K(self.inducing_points)
        Lu = torch.linalg.cholesky(Ku)  # Ku.cholesky()
        Ku_tiled = torch.tile(Ku[None, :, :], [self.num_outputs, 1, 1])
        Lu_tiled = torch.tile(Lu[None, :, :], [self.num_outputs, 1, 1])

        return Ku, Lu, Ku_tiled, Lu_tiled

    def conditional_ND(self, X, full_cov=False, y=None, ind_nan=None):
        """
        q(f|m, S; X, Z)
        @param X:
        @param full_cov:
        @param y:
        @param ind_nan:
        @return:
        """

        _, Lu, Ku_tiled, _ = self.get_prior_matrices()

        Kuf = self.kern(self.inducing_points, X)

        A = torch.linalg.solve_triangular(Lu, Kuf, upper=False)

        if not self.white:
            A = torch.linalg.solve_triangular(Lu.transpose(0, 1), A, upper=True)

        mean = A.transpose(0, 1).matmul(self.q_mu)

        A_tiled = A[None, :, :].repeat([self.num_outputs, 1, 1])
        I = torch.eye(self.num_inducing)[None, :, :].to(self.dtype).to(self.device)

        if self.white:
            SK = -I
        else:
            SK = -Ku_tiled

        if self.q_sqrt_tri is not None:
            qs = (
                torch.zeros((self.num_inducing, self.num_inducing))
                .to(self.dtype)
                .to(self.device)
            )
            li, lj = torch.tril_indices(self.num_inducing, self.num_inducing, 0)
            qs[li, lj] = self.q_sqrt_tri[0]
            q_sqrt = qs.unsqueeze(0)
            SK += q_sqrt.matmul(q_sqrt.transpose(1, 2))

        B = SK.matmul(A_tiled)

        if not full_cov:
            # (num_latent, num_X)
            delta_cov = (A_tiled * B).sum(1)
            Kff = self.kern.K_diag(X)

        var = Kff.unsqueeze(0) + delta_cov

        return mean + self.mean_function(X, ind_nan), var.transpose(0, 1)

    def KL(self):
        """
        KL DIVERGENCE COMPUTATION
        @return:
        """
        _, Lu, _, Lu_tiled = self.get_prior_matrices()

        D = self.num_outputs

        q_sqrt = (
            torch.zeros((self.num_inducing, self.num_inducing))
            .to(self.dtype)
            .to(self.device)
        )
        li, lj = torch.tril_indices(self.num_inducing, self.num_inducing, 0)
        q_sqrt[li, lj] = self.q_sqrt_tri[0].unsqueeze(0)

        # constant
        KL = -0.5 * D * self.num_inducing

        KL -= 0.5 * q_sqrt.diagonal(0, -2, -1).square().log().sum()

        if not self.white:
            KL += Lu.diagonal(0, -2, -1).log().sum() * D

            KL += (
                0.5
                * torch.linalg.solve_triangular(Lu_tiled, q_sqrt, upper=False)
                .square()
                .sum()
            )

            Kinv_m = torch.cholesky_solve(self.q_mu, Lu)
            KL += 0.5 * (self.q_mu * Kinv_m).sum()
        else:
            KL += 0.5 * q_sqrt.square().sum()
            KL += 0.5 * self.q_mu.square().sum()

        return KL
